import unittest
from ..support import BaseWebTest


class AdminViewTest(BaseWebTest, unittest.TestCase):

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['includes'] = 'kinto.plugins.admin'
        return settings

    def test_capability_is_exposed(self):
        self.maxDiff = None
        resp = self.app.get('/')
        capabilities = resp.json['capabilities']
        self.assertIn('admin', capabilities)
        self.assertIn('version', capabilities['admin'])
        del capabilities['admin']['version']
        expected = {
            "description": "Serves the admin console.",
            "url": ("https://github.com/Kinto/kinto-admin/"),
        }
        self.assertEqual(expected, capabilities['admin'])

    def test_admin_index_cat_be_reached(self):
        self.maxDiff = None
        resp = self.app.get('/admin/')
        assert "html" in resp.body.decode('utf-8')

    def test_admin_index_contains_authMethods_configurations(self):
        self.maxDiff = None
        resp = self.app.get('/admin/')
        body = resp.body.decode('utf-8')
        assert "authMethods" in body
        assert "basicauth" in body

    def test_admin_redirect_without_trailing_slash(self):
        resp = self.app.get('/admin', status=307)
        self.assertTrue(resp.headers['location'].endswith('/admin/'))



import mock
import webtest

from pyramid.config import Configurator
from pyramid.events import NewRequest
from pyramid.exceptions import ConfigurationError

import kinto.core
from kinto.core import initialization
from kinto.core.testing import unittest


class InitializationTest(unittest.TestCase):
    def test_fails_if_no_version_is_specified(self):
        config = Configurator()
        self.assertRaises(ConfigurationError, kinto.core.initialize, config)

    def test_uses_the_version_for_prefix(self):
        config = Configurator()
        kinto.core.initialize(config, '0.0.1', 'name')
        self.assertEqual(config.route_prefix, 'v0')

    def test_set_the_project_version_if_specified(self):
        config = Configurator()
        kinto.core.initialize(config, '0.0.1', 'name')
        self.assertEqual(config.registry.settings['project_version'], '0.0.1')

    def test_project_version_uses_setting_if_specified(self):
        config = Configurator(settings={'kinto.project_version': '1.0.0'})
        kinto.core.initialize(config, '0.0.1', 'name')
        self.assertEqual(config.registry.settings['project_version'], '1.0.0')

    def test_http_api_version_relies_on_project_version_by_default(self):
        config = Configurator()
        kinto.core.initialize(config, '0.1.0', 'name')
        self.assertEqual(config.registry.settings['http_api_version'], '0.1')

    def test_http_api_version_uses_setting_if_specified(self):
        config = Configurator(settings={'kinto.http_api_version': '1.3'})
        kinto.core.initialize(config, '0.0.1', 'name')
        self.assertEqual(config.registry.settings['http_api_version'], '1.3')

    def test_warns_if_project_name_is_empty(self):
        config = Configurator(settings={'kinto.project_name': ''})
        with mock.patch('kinto.core.initialization.warnings.warn') as mocked:
            kinto.core.initialize(config, '0.0.1')
            error_msg = 'No value specified for `project_name`'
            mocked.assert_any_call(error_msg)

    def test_warns_if_project_name_is_missing(self):
        config = Configurator()
        with mock.patch('kinto.core.initialization.warnings.warn') as mocked:
            kinto.core.initialize(config, '0.0.1')
            error_msg = 'No value specified for `project_name`'
            mocked.assert_any_call(error_msg)

    def test_set_the_project_name_if_specified(self):
        config = Configurator()
        kinto.core.initialize(config, '0.0.1', 'kinto')
        self.assertEqual(config.registry.settings['project_name'],
                         'kinto')

    def test_set_the_project_name_from_settings_even_if_specified(self):
        config = Configurator(settings={'kinto.project_name': 'kinto'})
        kinto.core.initialize(config, '0.0.1', 'readinglist')
        self.assertEqual(config.registry.settings['project_name'],
                         'kinto')

    def test_warns_if_not_https(self):
        with mock.patch('kinto.core.initialization.warnings.warn') as mocked:
            config = Configurator(settings={'project_name': 'kinto',
                                            'kinto.http_scheme': 'https'})
            kinto.core.initialize(config, '0.0.1')
            self.assertFalse(mocked.called)

            config = Configurator(settings={'kinto.http_scheme': 'http'})
            kinto.core.initialize(config, '0.0.1')
            error_msg = 'HTTPS is not enabled'
            mocked.assert_any_call(error_msg)

    def test_default_settings_are_overriden_if_specified_in_initialize(self):
        config = Configurator()
        defaults = {'kinto.paginate_by': 102}
        kinto.core.initialize(config, '0.0.1', 'project_name',
                              default_settings=defaults)
        self.assertEqual(config.registry.settings['paginate_by'], 102)

    def test_default_settings_are_overriden_by_application(self):
        config = Configurator(settings={'kinto.paginate_by': 10})
        kinto.core.initialize(config, '0.0.1', 'project_name')
        self.assertEqual(config.registry.settings['paginate_by'], 10)

    def test_specified_default_settings_are_overriden_by_application(self):
        config = Configurator(settings={'kinto.paginate_by': 5})
        defaults = {'kinto.paginate_by': 10}
        kinto.core.initialize(config, '0.0.1', 'project_name',
                              default_settings=defaults)
        self.assertEqual(config.registry.settings['paginate_by'], 5)

    def test_backends_are_not_instantiated_by_default(self):
        config = Configurator(settings={**kinto.core.DEFAULT_SETTINGS})
        kinto.core.initialize(config, '0.0.1', 'project_name')
        self.assertFalse(hasattr(config.registry, 'storage'))
        self.assertFalse(hasattr(config.registry, 'cache'))
        self.assertFalse(hasattr(config.registry, 'permission'))

    def test_backends_type_is_checked_when_instantiated(self):
        def config_fails(settings):
            config = Configurator(settings=settings)
            with self.assertRaises(ConfigurationError):
                kinto.core.initialize(config, '0.0.1', 'project_name')

        config_fails({'kinto.storage_backend': 'kinto.core.cache.memory'})
        config_fails({'kinto.cache_backend': 'kinto.core.storage.memory'})
        config_fails({'kinto.permission_backend': 'kinto.core.storage.memory'})

    def test_environment_values_override_configuration(self):
        import os

        envkey = 'KINTO_PROJECT_NAME'
        os.environ[envkey] = 'abc'

        config = Configurator(settings={'kinto.project_name': 'kinto'})
        kinto.core.initialize(config, '0.0.1')

        os.environ.pop(envkey)

        project_used = config.registry.settings['project_name']
        self.assertEqual(project_used, 'abc')


class ProjectSettingsTest(unittest.TestCase):
    def settings(self, provided):
        config = Configurator(settings=provided)
        kinto.core.initialize(config, '1.0', 'kinto')
        return config.get_settings()

    def test_uses_unprefixed_name(self):
        settings = {
            'paginate_by': 3.14
        }
        self.assertEqual(self.settings(settings)['paginate_by'], 3.14)

    def test_uses_project_name(self):
        settings = {
            'kinto.paginate_by': 42,
        }
        self.assertEqual(self.settings(settings)['paginate_by'], 42)

    def test_does_raise_valueerror_if_multiple_entries_are_equal(self):
        settings = {
            'paginate_by': 42,
            'kinto.paginate_by': 42,
        }
        self.settings(settings)  # Not raising.

    def test_does_raise_valueerror_if_entries_are_not_hashable(self):
        settings = {
            'events.listeners': ['Rémy', 42],
        }
        self.settings(settings)  # Not raising.

    def test_raises_valueerror_if_different_multiple_entries(self):
        settings = {
            'paginate_by': 42,
            'kinto.paginate_by': 3.14,
        }
        with self.assertRaises(ValueError):
            self.settings(settings)

    def test_environment_can_specify_project_name(self):
        import os

        envkey = 'KINTO_STORAGE_BACKEND'
        os.environ[envkey] = 'kinto_redis.storage'
        settings = {
            'kinto.storage_backend': 'kinto.core.storage.memory',
        }
        value = self.settings(settings)['storage_backend']
        os.environ.pop(envkey)
        self.assertEqual(value, 'kinto_redis.storage')

    def test_can_continue_to_use_cliquet_names(self):
        settings = {
            'kinto.permission_backend': 'cliquet.permission.memory'
        }
        with mock.patch('kinto.core.logger.warn') as mocked:
            new_settings = self.settings(settings)
            warning_message = ''.join([
                "Backend settings referring to cliquet are DEPRECATED. ",
                "Please update your kinto.permission_backend setting to ",
                "kinto.core.permission.memory ",
                "(was: cliquet.permission.memory).",
                ])
            mocked.assert_called_once_with(warning_message)
            self.assertEqual(new_settings['permission_backend'],
                             'kinto.core.permission.memory')


class ApplicationWrapperTest(unittest.TestCase):

    @unittest.skipIf(initialization.newrelic is None,
                     "newrelic is not installed.")
    @mock.patch('kinto.core.initialization.newrelic.agent')
    def test_newrelic_is_included_if_defined(self, mocked_newrelic):
        settings = {
            'newrelic_config': '/foo/bar.ini',
            'newrelic_env': 'test'
        }
        mocked_newrelic.WSGIApplicationWrapper.return_value = 'wrappedApp'
        app = kinto.core.install_middlewares(mock.sentinel.app, settings)
        mocked_newrelic.initialize.assert_called_with('/foo/bar.ini', 'test')
        self.assertEquals(app, 'wrappedApp')

    @unittest.skipIf(initialization.newrelic is None,
                     "newrelic is not installed.")
    @mock.patch('kinto.core.initialization.newrelic.agent')
    def test_newrelic_is_not_included_if_set_to_false(self, mocked_newrelic):
        settings = {'newrelic_config': False}
        app = kinto.core.install_middlewares(mock.sentinel.app, settings)
        mocked_newrelic.initialize.assert_not_called()
        self.assertEquals(app, mock.sentinel.app)

    @mock.patch('kinto.core.initialization.ProfilerMiddleware')
    def test_profiler_is_not_installed_if_set_to_false(self, mocked_profiler):
        settings = {'profiler_enabled': False}
        app = kinto.core.install_middlewares(mock.sentinel.app, settings)
        mocked_profiler.initialize.assert_not_called()
        self.assertEquals(app, mock.sentinel.app)

    @mock.patch('kinto.core.initialization.ProfilerMiddleware')
    def test_profiler_is_installed_if_set_to_true(self, mocked_profiler):
        settings = {
            'profiler_enabled': True,
            'profiler_dir': '/tmp/path'
        }
        mocked_profiler.return_value = 'wrappedApp'
        app = kinto.core.install_middlewares(mock.sentinel.app, settings)

        mocked_profiler.assert_called_with(
            mock.sentinel.app,
            restrictions='*kinto.core*',
            profile_dir='/tmp/path')

        self.assertEquals(app, 'wrappedApp')

    def test_load_default_settings_handle_prefix_attributes(self):
        config = mock.MagicMock()

        settings = {'project_name': 'myapp'}

        config.get_settings.return_value = settings
        initialization.load_default_settings(
            config, {'multiauth.policies': 'basicauth',
                     'kinto.http_scheme': 'https',
                     'myapp.http_host': 'localhost:8888'})

        self.assertIn('multiauth.policies', settings)
        self.assertNotIn('policies', settings)
        self.assertEqual(settings['multiauth.policies'], 'basicauth')

        self.assertIn('http_scheme', settings)
        self.assertNotIn('kinto.http_scheme', settings)
        self.assertEqual(settings['http_scheme'], 'https')

        self.assertIn('http_host', settings)
        self.assertNotIn('myapp.http_host', settings)
        self.assertEqual(settings['http_host'], 'localhost:8888')


class StatsDConfigurationTest(unittest.TestCase):
    def setUp(self):
        settings = {**kinto.core.DEFAULT_SETTINGS, 'statsd_url': 'udp://host:8080'}
        self.config = Configurator(settings=settings)
        self.config.registry.storage = {}
        self.config.registry.cache = {}
        self.config.registry.permission = {}

        patch = mock.patch('kinto.core.statsd.load_from_config')
        self.mocked = patch.start()
        self.addCleanup(patch.stop)

    def test_statsd_isnt_called_if_statsd_url_is_not_set(self):
        self.config.add_settings({
            'statsd_url': None
        })
        initialization.setup_statsd(self.config)
        self.mocked.assert_not_called()

    def test_statsd_is_set_to_none_if_statsd_url_not_set(self):
        self.config.add_settings({
            'statsd_url': None
        })
        initialization.setup_statsd(self.config)
        self.assertEqual(self.config.registry.statsd, None)

    def test_statsd_is_called_if_statsd_url_is_set(self):
        initialization.setup_statsd(self.config)
        self.mocked.assert_called_with(self.config)
        # See `tests/core/test_statsd.py` for instantiation tests.

    def test_statsd_is_expose_in_the_registry_if_url_is_set(self):
        initialization.setup_statsd(self.config)
        self.assertEqual(self.config.registry.statsd, self.mocked.return_value)

    def test_statsd_is_set_on_cache(self):
        c = initialization.setup_statsd(self.config)
        c.watch_execution_time.assert_any_call({}, prefix='backend')

    def test_statsd_is_set_on_storage(self):
        c = initialization.setup_statsd(self.config)
        c.watch_execution_time.assert_any_call({}, prefix='backend')

    def test_statsd_is_set_on_permission(self):
        c = initialization.setup_statsd(self.config)
        c.watch_execution_time.assert_any_call({}, prefix='backend')

    def test_statsd_is_set_on_authentication(self):
        c = initialization.setup_statsd(self.config)
        c.watch_execution_time.assert_any_call(None, prefix='authentication')

    def test_statsd_counts_nothing_on_anonymous_requests(self):
        kinto.core.initialize(self.config, '0.0.1', 'project_name')
        app = webtest.TestApp(self.config.make_wsgi_app())
        app.get('/')
        self.assertFalse(self.mocked.count.called)

    def test_statsd_counts_views_and_methods(self):
        kinto.core.initialize(self.config, '0.0.1', 'project_name')
        app = webtest.TestApp(self.config.make_wsgi_app())
        app.get('/v0/__heartbeat__')
        self.mocked().count.assert_any_call('view.heartbeat.GET')

    def test_statsd_counts_unknown_urls(self):
        kinto.core.initialize(self.config, '0.0.1', 'project_name')
        app = webtest.TestApp(self.config.make_wsgi_app())
        app.get('/v0/coucou', status=404)
        self.assertFalse(self.mocked.count.called)

    @mock.patch('kinto.core.utils.hmac_digest')
    def test_statsd_counts_unique_users(self, digest_mocked):
        digest_mocked.return_value = 'mat'
        kinto.core.initialize(self.config, '0.0.1', 'project_name')
        app = webtest.TestApp(self.config.make_wsgi_app())
        headers = {'Authorization': 'Basic bWF0Og=='}
        app.get('/v0/', headers=headers)
        self.mocked().count.assert_any_call('users', unique='basicauth:mat')

    def test_statsd_counts_authentication_types(self):
        kinto.core.initialize(self.config, '0.0.1', 'project_name')
        app = webtest.TestApp(self.config.make_wsgi_app())
        headers = {'Authorization': 'Basic bWF0Og=='}
        app.get('/v0/', headers=headers)
        self.mocked().count.assert_any_call('authn_type.basicauth')


class RequestsConfigurationTest(unittest.TestCase):
    def _get_app(self, settings={}):
        app_settings = {
            'storage_backend': 'kinto.core.storage.memory',
            'cache_backend': 'kinto_redis.cache',
        }
        app_settings.update(**settings)
        config = Configurator(settings=app_settings)
        kinto.core.initialize(config, '0.0.1', 'name')
        return webtest.TestApp(config.make_wsgi_app())

    def test_requests_have_a_bound_data_attribute(self):
        config = Configurator()
        kinto.core.initialize(config, '0.0.1', 'name')

        def on_new_request(event):
            data = event.request.bound_data
            self.assertEqual(data, {})
            self.assertEqual(id(data), id(event.request.bound_data))

        config.add_subscriber(on_new_request, NewRequest)
        app = webtest.TestApp(config.make_wsgi_app())
        app.get('/v0/')

    def test_subrequests_share_parent_bound_data(self):
        config = Configurator()
        kinto.core.initialize(config, '0.0.1', 'name')

        bound_datas = set()

        def on_new_request(event):
            bound_datas.add(id(event.request.bound_data))

        config.add_subscriber(on_new_request, NewRequest)
        app = webtest.TestApp(config.make_wsgi_app())
        app.post_json('/v0/batch', {'requests': [{'path': '/'}]})
        self.assertEqual(len(bound_datas), 1)

    def test_by_default_relies_on_pyramid_application_url(self):
        app = self._get_app()
        resp = app.get('/v0/')
        self.assertEqual(resp.json['url'], 'http://localhost/v0/')

    def test_by_default_relies_on_incoming_headers(self):
        app = self._get_app()
        resp = app.get('/v0/', headers={'Host': 'server:8888'})
        self.assertEqual(resp.json['url'], 'http://server:8888/v0/')

    def test_by_default_relies_on_wsgi_environment(self):
        app = self._get_app()
        environ = {
            'wsgi.url_scheme': 'https',
            'HTTP_HOST': 'server:44311'
        }
        resp = app.get('/v0/', extra_environ=environ)
        self.assertEqual(resp.json['url'], 'https://server:44311/v0/')

    def test_http_scheme_overrides_the_wsgi_environment(self):
        app = self._get_app({'http_scheme': 'http2'})
        environ = {
            'wsgi.url_scheme': 'https'
        }
        resp = app.get('/v0/', extra_environ=environ)
        self.assertEqual(resp.json['url'], 'http2://localhost:80/v0/')

    def test_http_host_overrides_the_wsgi_environment(self):
        app = self._get_app({'http_host': 'server'})
        environ = {
            'HTTP_HOST': 'elb:44311'
        }
        resp = app.get('/v0/', extra_environ=environ)
        self.assertEqual(resp.json['url'], 'http://server/v0/')

    def test_http_host_overrides_the_request_headers(self):
        app = self._get_app({'http_host': 'server'})
        resp = app.get('/v0/', headers={'Host': 'elb:8888'})
        self.assertEqual(resp.json['url'], 'http://server/v0/')


class PluginsTest(unittest.TestCase):
    def test_kinto_core_includes_are_included_manually(self):
        config = Configurator(settings={**kinto.core.DEFAULT_SETTINGS})
        config.add_settings({'includes': 'elastic history'})
        config.route_prefix = 'v2'

        with mock.patch.object(config, 'include'):
            with mock.patch.object(config, 'scan'):
                kinto.core.includeme(config)

                config.include.assert_any_call('elastic')
                config.include.assert_any_call('history')

    def make_app(self):
        config = Configurator(settings={**kinto.core.DEFAULT_SETTINGS})
        config.add_settings({
            'permission_backend': 'kinto.core.permission.memory',
            'includes': 'tests.core.testplugin'
        })
        kinto.core.initialize(config, '0.0.1', 'name')
        return webtest.TestApp(config.make_wsgi_app())

    def test_plugin_can_define_protected_views(self):
        app = self.make_app()
        app.post('/v0/attachment', status=401)
        headers = {'Authorization': 'Basic bWF0OjE='}
        app.post('/v0/attachment', headers=headers, status=403)

    def test_plugin_benefits_from_cors_setup(self):
        app = self.make_app()
        headers = {
            'Origin': 'lolnet.org',
            'Access-Control-Request-Method': 'POST'
        }
        resp = app.options('/v0/attachment', headers=headers, status=200)
        self.assertIn('Access-Control-Allow-Origin', resp.headers)



import unittest

from kinto.core.testing import get_user_headers

from .support import (BaseWebTest, MINIMALIST_BUCKET,
                      MINIMALIST_COLLECTION, MINIMALIST_RECORD)


class CollectionViewTest(BaseWebTest, unittest.TestCase):

    collections_url = '/buckets/beers/collections'
    collection_url = '/buckets/beers/collections/barley'

    def setUp(self):
        super().setUp()
        self.app.put_json('/buckets/beers', MINIMALIST_BUCKET,
                          headers=self.headers)
        resp = self.app.put_json(self.collection_url,
                                 MINIMALIST_COLLECTION,
                                 headers=self.headers)
        self.record = resp.json['data']

    def test_collection_endpoint_lists_them_all(self):
        resp = self.app.get(self.collections_url, headers=self.headers)
        records = resp.json['data']
        self.assertEqual(len(records), 1)
        self.assertEqual(records[0]['id'], 'barley')

    def test_collections_can_be_put_with_simple_name(self):
        self.assertEqual(self.record['id'], 'barley')

    def test_collections_name_should_be_simple(self):
        self.app.put_json('/buckets/beers/collections/__barley__',
                          MINIMALIST_COLLECTION,
                          headers=self.headers,
                          status=400)

    def test_collections_should_reject_unaccepted_request_content_type(self):
        headers = {**self.headers, 'Content-Type': 'text/plain'}
        self.app.put('/buckets/beers/collections/barley',
                     MINIMALIST_COLLECTION,
                     headers=headers,
                     status=415)

    def test_unknown_bucket_raises_403(self):
        other_bucket = self.collections_url.replace('beers', 'sodas')
        self.app.get(other_bucket, headers=self.headers, status=403)

    def test_unknown_collection_raises_404(self):
        other_collection = self.collection_url.replace('barley', 'pills')
        resp = self.app.get(other_collection, headers=self.headers, status=404)
        self.assertEqual(resp.json['details']['id'], 'pills')
        self.assertEqual(resp.json['details']['resource_name'], 'collection')

    def test_collections_are_isolated_by_bucket(self):
        other_bucket = self.collection_url.replace('beers', 'sodas')
        self.app.put_json('/buckets/sodas',
                          MINIMALIST_BUCKET,
                          headers=self.headers)
        self.app.get(other_bucket, headers=self.headers, status=404)

    def test_create_permissions_can_be_added_on_collections(self):
        collection = {**MINIMALIST_COLLECTION, 'permissions': {'record:create': ['fxa:user']}}
        resp = self.app.put_json('/buckets/beers/collections/barley',
                                 collection,
                                 headers=self.headers,
                                 status=200)
        permissions = resp.json['permissions']
        self.assertIn('fxa:user', permissions['record:create'])

    def test_wrong_create_permissions_cannot_be_added_on_collections(self):
        collection = {**MINIMALIST_COLLECTION, 'permissions': {'collection:create': ['fxa:user']}}
        self.app.put_json('/buckets/beers/collections/barley',
                          collection,
                          headers=self.headers,
                          status=400)

    def test_collections_can_handle_arbitrary_attributes(self):
        fingerprint = "5866f245a00bb3a39100d31b2f14d453"
        collection = {**MINIMALIST_COLLECTION, 'data': {'fingerprint': fingerprint}}
        resp = self.app.put_json('/buckets/beers/collections/barley',
                                 collection,
                                 headers=self.headers,
                                 status=200)
        data = resp.json['data']
        self.assertIn('fingerprint', data)
        self.assertEqual(data['fingerprint'], fingerprint)

    def test_collections_can_be_filtered_by_arbitrary_attribute(self):
        collection = {**MINIMALIST_COLLECTION, 'data': {'size': 3}}
        self.app.put_json('/buckets/beers/collections/moderator',
                          collection,
                          headers=self.headers)
        resp = self.app.get('/buckets/beers/collections?min_size=2',
                            headers=self.headers)
        data = resp.json['data']
        self.assertEqual(len(data), 1)


class CollectionDeletionTest(BaseWebTest, unittest.TestCase):

    collection_url = '/buckets/beers/collections/barley'

    def setUp(self):
        super().setUp()
        bucket = {**MINIMALIST_BUCKET,
                  'permissions': {'collection:create': ['system.Everyone'],
                                  'read': ['system.Everyone']}}
        self.app.put_json('/buckets/beers', bucket,
                          headers=self.headers)
        self.app.put_json(self.collection_url, MINIMALIST_COLLECTION,
                          headers=self.headers)
        r = self.app.post_json(self.collection_url + '/records',
                               MINIMALIST_RECORD,
                               headers=self.headers)
        record_id = r.json['data']['id']
        self.record_url = self.collection_url + '/records/{}'.format(record_id)
        self.app.delete(self.collection_url, headers=self.headers)

    def test_collections_can_be_deleted(self):
        self.app.get(self.collection_url, headers=self.headers, status=404)

    def test_collections_can_be_deleted_in_bulk(self):
        alice_headers = get_user_headers('alice')
        self.app.put_json('/buckets/beers/collections/1',
                          MINIMALIST_COLLECTION, headers=self.headers)
        self.app.put_json('/buckets/beers/collections/2',
                          MINIMALIST_COLLECTION, headers=alice_headers)
        self.app.put_json('/buckets/beers/collections/3',
                          MINIMALIST_COLLECTION, headers=alice_headers)
        self.app.delete('/buckets/beers/collections',
                        headers=alice_headers)
        resp = self.app.get('/buckets/beers/collections', headers=self.headers)
        self.assertEqual(len(resp.json['data']), 1)

    def test_records_of_collection_are_deleted_too(self):
        self.app.put_json(self.collection_url, MINIMALIST_COLLECTION,
                          headers=self.headers)
        self.app.get(self.record_url, headers=self.headers, status=404)

        # Verify tombstones
        resp = self.app.get('{}/records?_since=0'.format(self.collection_url),
                            headers=self.headers)
        self.assertEqual(len(resp.json['data']), 0)

    def test_can_be_created_after_deletion_with_if_none_match_star(self):
        headers = {**self.headers, 'If-None-Match': '*'}
        self.app.put_json(self.collection_url, MINIMALIST_COLLECTION,
                          headers=headers, status=201)


class CollectionCreationTest(BaseWebTest, unittest.TestCase):

    collections_url = '/buckets/beers/collections'

    def setUp(self):
        super().setUp()
        self.app.put_json('/buckets/beers', MINIMALIST_BUCKET,
                          headers=self.headers)

    def test_collection_can_be_created_with_post(self):
        r = self.app.post_json(self.collections_url,
                               MINIMALIST_COLLECTION,
                               headers=self.headers)
        self.assertEqual(r.status_code, 201)
        self.assertTrue(len(r.json['data']['id']) == 8)

    def test_collection_can_be_specified_in_post(self):
        collection = 'barley'
        r = self.app.post_json(self.collections_url,
                               {'data': {'id': collection}},
                               headers=self.headers)
        self.assertEqual(r.status_code, 201)
        self.assertEqual(r.json['data']['id'], collection)

    def test_collection_already_exists_post(self):
        collection = "barley"
        self.app.post_json(self.collections_url,
                           {'data': {'id': collection}},
                           headers=self.headers)
        r = self.app.post_json(self.collections_url,
                               {'data': {'id': collection}},
                               headers=self.headers)
        self.assertEqual(r.json['data']['id'], collection)
        self.assertEqual(r.status_code, 200)



import mock
import os
import uuid
from pyramid import testing

from kinto.core import initialization
from kinto.core.events import ResourceChanged, ResourceRead, ACTIONS
from kinto.core.listeners import ListenerBase
from kinto.core.testing import unittest


UID = str(uuid.uuid4())


class ViewSet:
    def get_name(*args, **kw):
        return 'collection'


class Service:
    viewset = ViewSet()


class Match:
    cornice_services = {'watev': Service()}
    pattern = 'watev'


class Request:
    path = '/1/bucket/collection/'
    prefixed_userid = 'tarek'
    matchdict = {'id': UID}
    registry = matched_route = Match()
    current_resource_name = 'bucket'


class ListenerSetupTest(unittest.TestCase):
    def setUp(self):
        demo_patch = mock.patch('tests.core.listeners.load_from_config')
        self.addCleanup(demo_patch.stop)
        self.demo_mocked = demo_patch.start()

    def make_app(self, extra_settings={}):
        settings = {
            'event_listeners': 'tests.core.listeners',
        }
        settings.update(**extra_settings)
        config = testing.setUp(settings=settings)
        config.commit()
        initialization.setup_listeners(config)
        return config

    def test_listener_module_is_specified_via_settings(self):
        self.make_app({
            'event_listeners': 'demo',
            'event_listeners.demo.use': 'tests.core.listeners',
        })
        self.assertTrue(self.demo_mocked.called)

    def test_listener_module_can_be_specified_via_listeners_list(self):
        self.make_app({
            'event_listeners': 'demo',
            'event_listeners.demo.use': 'tests.core.listeners',
        })
        self.assertTrue(self.demo_mocked.called)

    def test_callback_called_when_action_is_not_filtered(self):
        config = self.make_app({
            'event_listeners': 'demo',
            'event_listeners.demo.use': 'tests.core.listeners',
        })
        ev = ResourceChanged({'action': ACTIONS.CREATE.value}, [], Request())
        config.registry.notify(ev)

        self.assertTrue(self.demo_mocked.return_value.called)

    def test_callback_is_not_called_when_action_is_filtered(self):
        config = self.make_app({
            'event_listeners': 'demo',
            'event_listeners.demo.use': 'tests.core.listeners',
            'event_listeners.demo.actions': 'delete',
        })
        ev = ResourceChanged({'action': ACTIONS.CREATE.value}, [], Request())
        config.registry.notify(ev)

        self.assertFalse(self.demo_mocked.return_value.called)

    def test_callback_called_when_resource_is_not_filtered(self):
        config = self.make_app({
            'event_listeners': 'demo',
            'event_listeners.demo.use': 'tests.core.listeners',
        })
        event = ResourceChanged({'action': ACTIONS.CREATE.value,
                                 'resource_name': 'mushroom'}, [], Request())
        config.registry.notify(event)

        self.assertTrue(self.demo_mocked.return_value.called)

    def test_callback_is_not_called_when_resource_is_filtered(self):
        config = self.make_app({
            'event_listeners': 'demo',
            'event_listeners.demo.use': 'tests.core.listeners',
            'event_listeners.demo.resources': 'toad',
        })
        event = ResourceChanged({'action': ACTIONS.CREATE.value,
                                 'resource_name': 'mushroom'}, [], Request())
        config.registry.notify(event)

        self.assertFalse(self.demo_mocked.return_value.called)

    def test_callback_is_not_called_on_read_by_default(self):
        config = self.make_app({
            'event_listeners': 'demo',
            'event_listeners.demo.use': 'tests.core.listeners',
        })
        event = ResourceRead({'action': ACTIONS.READ.value}, [], Request())
        config.registry.notify(event)

        self.assertFalse(self.demo_mocked.return_value.called)

    def test_callback_is_called_on_read_if_specified(self):
        config = self.make_app({
            'event_listeners': 'demo',
            'event_listeners.demo.use': 'tests.core.listeners',
            'event_listeners.demo.actions': 'read',
        })
        event = ResourceRead({'action': ACTIONS.READ.value}, [], Request())
        config.registry.notify(event)

        self.assertTrue(self.demo_mocked.return_value.called)

    def test_same_callback_is_called_for_read_and_write_specified(self):
        config = self.make_app({
            'event_listeners': 'demo',
            'event_listeners.demo.use': 'tests.core.listeners',
            'event_listeners.demo.actions': 'read create delete',
        })
        ev = ResourceRead({'action': ACTIONS.READ.value}, [], Request())
        config.registry.notify(ev)
        ev = ResourceChanged({'action': ACTIONS.CREATE.value}, [], Request())
        config.registry.notify(ev)

        self.assertEqual(self.demo_mocked.return_value.call_count, 2)

    def test_loading_can_read_configuration_from_environment(self):
        environ = {
            "KINTO_EVENT_LISTENERS": "kvstore",
            "KINTO_EVENT_LISTENERS_KVSTORE_USE": "tests.core.listeners",
            "KINTO_EVENT_LISTENERS_KVSTORE_URL": "demo://demo:6379/0",
            "KINTO_EVENT_LISTENERS_KVSTORE_POOL_SIZE": "5",
            "KINTO_EVENT_LISTENERS_KVSTORE_LISTNAME": "queue",
            "KINTO_EVENT_LISTENERS_KVSTORE_ACTIONS": "delete",
            "KINTO_EVENT_LISTENERS_KVSTORE_RESOURCES": "toad",
        }
        os.environ.update(**environ)

        config = self.make_app({
            # With real/full initialization, these should not be necessary:
            'project_name': 'kinto',
            'event_listeners': 'kvstore'
        })

        # Listener is instantiated.
        self.assertTrue(self.demo_mocked.called)

        # Action filtering is read from ENV.
        event = ResourceChanged({'action': ACTIONS.DELETE.value,
                                 'resource_name': 'toad'}, [], Request())
        config.registry.notify(event)
        self.assertTrue(self.demo_mocked.return_value.called)

        self.demo_mocked.reset_mock()

        # Action filtering is read from ENV.
        event = ResourceChanged({'action': ACTIONS.CREATE.value},
                                [], Request())
        config.registry.notify(event)
        self.assertFalse(self.demo_mocked.return_value.called)

        # Resource filtering is read from ENV.
        event = ResourceChanged({'action': ACTIONS.CREATE.value,
                                 'resource_name': 'mushroom'}, [], Request())
        config.registry.notify(event)
        self.assertFalse(self.demo_mocked.return_value.called)

        # Clean-up.
        for k in environ.keys():
            os.environ.pop(k)


class ListenerBaseTest(unittest.TestCase):

    def test_not_implemented(self):
        # make sure we can't use the base listener
        listener = ListenerBase()
        self.assertRaises(NotImplementedError, listener, object())



import mock
import os
import json
import tempfile
import unittest

from kinto.core.views.version import version_view
from .support import BaseWebTest


class VersionViewTest(BaseWebTest, unittest.TestCase):
    def setUp(self):
        try:
            delattr(version_view, '__json__')  # Clean-up memoization.
        except AttributeError:
            pass

    tearDown = setUp

    def test_return_the_version_file_in_current_folder_if_present(self):
        content = {'version': '0.8.1'}
        fake_file = mock.mock_open(read_data=json.dumps(content))
        with mock.patch('os.path.exists'):
            with mock.patch('kinto.core.views.version.open', fake_file, create=True):
                response = self.app.get('/__version__')
                assert response.json == content

    def test_return_a_404_if_version_file_if_not_present(self):
        self.app.get('/__version__', status=404)

    def test_return_the_version_file_specified_in_setting_if_present(self):
        custom_path = tempfile.mktemp()
        content = {'foo': 'lala'}
        with open(custom_path, 'w') as f:
            json.dump(content, f)
        self.addCleanup(lambda: os.remove(custom_path))

        with mock.patch.dict(self.app.app.registry.settings,
                             [('version_json_path', custom_path)]):
            response = self.app.get('/__version__')

        assert response.json == content



import colander
import mock
import uuid
import unittest

from pyramid.response import Response

from kinto.core.views.batch import BatchPayloadSchema, batch as batch_service
from kinto.core.testing import DummyRequest
from kinto.core.utils import json

from .support import BaseWebTest


class BatchViewTest(BaseWebTest, unittest.TestCase):

    def test_does_not_require_authentication(self):
        body = {'requests': []}
        self.app.post_json('/batch', body)

    def test_returns_400_if_body_has_missing_requests(self):
        self.app.post('/batch', {}, headers=self.headers, status=400)

    def test_returns_responses_if_schema_is_valid(self):
        body = {'requests': []}
        resp = self.app.post_json('/batch', body, headers=self.headers)
        self.assertIn('responses', resp.json)

    def test_defaults_are_applied_to_requests(self):
        request = {'path': '/v0/'}
        defaults = {'method': 'POST'}
        result = self.app.post_json('/batch',
                                    {'requests': [request],
                                     'defaults': defaults})
        self.assertEqual(result.json['responses'][0]['status'], 405)

    def test_only_post_is_allowed(self):
        self.app.get('/batch', headers=self.headers, status=405)
        self.app.put('/batch', headers=self.headers, status=405)
        self.app.patch('/batch', headers=self.headers, status=405)
        self.app.delete('/batch', headers=self.headers, status=405)

    def test_batch_adds_missing_api_with_prefix(self):
        request = {'path': '/v0/'}
        body = {'requests': [request]}
        resp = self.app.post_json('/batch', body, headers=self.headers)
        hello = resp.json['responses'][0]
        self.assertEqual(hello['path'], '/v0/')
        self.assertEqual(hello['status'], 200)
        self.assertEqual(hello['body']['project_name'], 'myapp')
        self.assertIn('application/json', hello['headers']['Content-Type'])

    def test_empty_response_body_with_head(self):
        request = {'path': '/v0/', 'method': 'HEAD'}
        body = {'requests': [request]}
        resp = self.app.post_json('/batch', body, headers=self.headers)
        head = resp.json['responses'][0]
        self.assertEqual(head['body'], '')
        self.assertNotEqual(len(head['headers']), 0)

    def test_api_errors_are_json_formatted(self):
        request = {'path': '/unknown'}
        body = {'requests': [request]}
        resp = self.app.post_json('/batch', body, headers=self.headers)
        error = resp.json['responses'][0]
        self.assertEqual(error['body']['code'], 404)

    def test_internal_errors_makes_the_batch_fail(self):
        request = {'path': '/v0/'}
        body = {'requests': [request]}

        with mock.patch('kinto.core.views.hello.get_eos') as mocked:
            mocked.side_effect = AttributeError
            self.app.post_json('/batch', body, headers=self.headers,
                               status=500)

    def test_errors_handled_by_view_does_not_make_the_batch_fail(self):
        from requests.exceptions import HTTPError

        request = {'path': '/v0/'}
        body = {'requests': [request]}

        with mock.patch('kinto.core.views.hello.get_eos') as mocked:
            response = mock.MagicMock(status_code=404)
            mocked.side_effect = HTTPError(response=response)
            resp = self.app.post_json('/batch', body, headers=self.headers,
                                      status=200)
            subresponse = resp.json['responses'][0]['body']
            self.assertEqual(subresponse, {
                'errno': 999,
                'code': 404,
                'error': 'Not Found'
            })

    def test_batch_cannot_be_recursive(self):
        requests = {'requests': [{'path': '/v0/'}]}
        request = {'method': 'POST', 'path': '/v0/batch', 'body': requests}
        body = {'requests': [request]}
        resp = self.app.post_json('/batch', body, status=400)
        self.assertIn('Recursive', resp.json['message'])

    def test_batch_validates_json(self):
        body = """{"requests": [{"path": "/v0/"},]}"""
        resp = self.app.post('/batch', body, status=400,
                             headers={'Content-Type': 'application/json'})
        self.assertIn('Invalid JSON', resp.json['message'])

    def test_responses_are_resolved_with_api_with_prefix(self):
        request = {'path': '/'}
        body = {'requests': [request]}
        resp = self.app.post_json('/batch', body, headers=self.headers)
        hello = resp.json['responses'][0]
        self.assertEqual(hello['path'], '/v0/')
        self.assertEqual(hello['status'], 200)
        self.assertEqual(hello['body']['project_name'], 'myapp')
        self.assertIn('application/json', hello['headers']['Content-Type'])

    def test_redirect_responses_are_followed(self):
        request = {'path': '/mushrooms/'}  # trailing slash
        body = {'requests': [request]}
        resp = self.app.post_json('/batch', body, headers=self.headers)
        collection = resp.json['responses'][0]
        self.assertEqual(collection['status'], 200)
        self.assertEqual(collection['path'], '/v0/mushrooms')
        self.assertEqual(collection['body'], {'data': []})

    def test_body_is_transmitted_during_redirect(self):
        request = {
            'method': 'PUT',
            'path': '/mushrooms/{}/'.format(str(uuid.uuid4())),
            'body': {'data': {'name': 'Trompette de la mort'}}
        }
        body = {'requests': [request]}
        resp = self.app.post_json('/batch', body, headers=self.headers)
        response = resp.json['responses'][0]
        self.assertEqual(response['status'], 201)
        record = response['body']['data']
        self.assertEqual(record['name'], 'Trompette de la mort')

    def test_400_error_message_is_forwarded(self):
        headers = {**self.headers, 'If-Match': '"*"'}
        request = {
            'method': 'PUT',
            'path': '/mushrooms/{}'.format(str(uuid.uuid4())),
            'body': {'data': {'name': 'Trompette de la mort'}},
            'headers': headers
        }
        body = {'requests': [request, request]}
        resp = self.app.post_json('/batch', body, status=200)
        self.assertEqual(resp.json['responses'][1]['status'], 400)
        msg = 'If-Match in header: The value should be integer between double quotes.'
        self.assertEqual(resp.json['responses'][1]['body']['message'], msg)

    def test_412_errors_are_forwarded(self):
        headers = {**self.headers, 'If-None-Match': '*'}
        request = {
            'method': 'PUT',
            'path': '/mushrooms/{}'.format(str(uuid.uuid4())),
            'body': {'data': {'name': 'Trompette de la mort'}},
            'headers': headers
        }
        body = {'requests': [request, request]}
        resp = self.app.post_json('/batch', body, status=200)
        self.assertEqual(resp.json['responses'][0]['status'], 201)
        self.assertEqual(resp.json['responses'][1]['status'], 412)


class BatchSchemaTest(unittest.TestCase):
    def setUp(self):
        self.schema = BatchPayloadSchema()

    def assertInvalid(self, payload):
        self.assertRaises(colander.Invalid, self.schema.deserialize, payload)

    def test_requests_is_mandatory(self):
        self.assertInvalid({})

    def test_unknown_attributes_are_dropped(self):
        deserialized = self.schema.deserialize({'requests': [], 'unknown': 42})
        self.assertNotIn('unknown', deserialized)

    def test_list_of_requests_can_be_empty(self):
        self.schema.deserialize({'requests': []})

    def test_list_of_requests_must_be_a_list(self):
        self.assertInvalid({'requests': {}})

    def test_list_of_requests_must_be_dicts(self):
        request = 42
        self.assertInvalid({'defaults': {'path': '/'}, 'requests': [request]})

    def test_request_path_must_start_with_slash(self):
        request = {'path': 'http://localhost'}
        self.assertInvalid({'requests': [request]})

    def test_request_path_is_mandatory(self):
        request = {'method': 'HEAD'}
        self.assertInvalid({'requests': [request]})

    def test_request_method_must_be_known_uppercase_word(self):
        request = {'path': '/', 'method': 'get'}
        self.assertInvalid({'requests': [request]})

    #
    # headers
    #

    def test_request_headers_should_be_strings(self):
        headers = {'Accept': 3.14}
        request = {'path': '/', 'headers': headers}
        self.assertInvalid({'requests': [request]})

    def test_request_headers_cannot_be_recursive(self):
        headers = {'Accept': {'sub': 'dict'}}
        request = {'path': '/', 'headers': headers}
        self.assertInvalid({'requests': [request]})

    def test_request_headers_are_preserved(self):
        headers = {'Accept': 'audio/*'}
        request = {'path': '/', 'headers': headers}
        deserialized = self.schema.deserialize({'requests': [request]})
        self.assertEqual(deserialized['requests'][0]['headers']['Accept'],
                         'audio/*')

    #
    # body
    #

    def test_body_is_an_arbitrary_mapping(self):
        payload = {"json": "payload"}
        request = {'path': '/', 'body': payload}
        deserialized = self.schema.deserialize({'requests': [request]})
        self.assertEqual(deserialized['requests'][0]['body'], payload)

    #
    # defaults
    #

    def test_defaults_must_be_a_mapping_if_specified(self):
        request = {'path': '/'}
        batch_payload = {'requests': [request], 'defaults': 42}
        self.assertInvalid(batch_payload)

    def test_defaults_must_be_a_request_schema_if_specified(self):
        request = {'path': '/'}
        defaults = {'body': 3}
        batch_payload = {'requests': [request], 'defaults': defaults}
        self.assertInvalid(batch_payload)

    def test_unknown_defaults_are_ignored_silently(self):
        request = {'path': '/'}
        defaults = {'foo': 'bar'}
        batch_payload = {'requests': [request], 'defaults': defaults}
        result = self.schema.deserialize(batch_payload)
        self.assertNotIn('foo', result['requests'][0])

    def test_defaults_can_be_specified_empty(self):
        request = {'path': '/'}
        defaults = {}
        batch_payload = {'requests': [request], 'defaults': defaults}
        self.schema.deserialize(batch_payload)

    def test_defaults_path_is_applied_to_requests(self):
        request = {'method': 'GET'}
        defaults = {'path': '/'}
        batch_payload = {'requests': [request], 'defaults': defaults}
        result = self.schema.deserialize(batch_payload)
        self.assertEqual(result['requests'][0]['path'], '/')

    def test_defaults_body_is_applied_to_requests(self):
        request = {'path': '/'}
        defaults = {'body': {'json': 'payload'}}
        batch_payload = {'requests': [request], 'defaults': defaults}
        result = self.schema.deserialize(batch_payload)
        self.assertEqual(result['requests'][0]['body'], {'json': 'payload'})

    def test_defaults_headers_are_applied_to_requests(self):
        request = {'path': '/'}
        defaults = {'headers': {'Content-Type': 'text/html'}}
        batch_payload = {'requests': [request], 'defaults': defaults}
        result = self.schema.deserialize(batch_payload)
        self.assertEqual(result['requests'][0]['headers']['Content-Type'],
                         'text/html')

    def test_defaults_values_do_not_overwrite_requests_values(self):
        request = {'path': '/', 'headers': {'Authorization': 'me'}}
        defaults = {'headers': {'Authorization': 'you', 'Accept': '*/*'}}
        batch_payload = {'requests': [request], 'defaults': defaults}
        result = self.schema.deserialize(batch_payload)
        self.assertEqual(result['requests'][0]['headers'],
                         {'Authorization': 'me', 'Accept': '*/*'})

    def test_defaults_values_for_path_must_start_with_slash(self):
        request = {}
        defaults = {'path': 'http://localhost'}
        batch_payload = {'requests': [request], 'defaults': defaults}
        self.assertInvalid(batch_payload)


class BatchServiceTest(unittest.TestCase):
    def setUp(self):
        self.method, self.view, self.options = batch_service.definitions[0]
        self.request = DummyRequest()

    def post(self, validated):
        self.request.validated = {'body': validated}
        return self.view(self.request)

    def test_returns_empty_list_of_responses_if_requests_empty(self):
        result = self.post({'requests': []})
        self.assertEqual(result['responses'], [])

    def test_returns_one_response_per_request(self):
        requests = [{'path': '/'}]
        result = self.post({'requests': requests})
        self.assertEqual(len(result['responses']), len(requests))

    def test_relies_on_pyramid_invoke_subrequest(self):
        self.post({'requests': [{'path': '/'}]})
        self.assertTrue(self.request.invoke_subrequest.called)

    def test_returns_requests_path_in_responses(self):
        result = self.post({'requests': [{'path': '/'}]})
        self.assertEqual(result['responses'][0]['path'], '/v0/')

    def test_subrequests_have_parent_attribute(self):
        self.request.path = '/batch'
        self.post({'requests': [{'path': '/'}]})
        subrequest, = self.request.invoke_subrequest.call_args[0]
        self.assertEqual(subrequest.parent.path, '/batch')

    def test_subrequests_are_GET_by_default(self):
        self.post({'requests': [{'path': '/'}]})
        subrequest, = self.request.invoke_subrequest.call_args[0]
        self.assertEqual(subrequest.method, 'GET')

    def test_original_request_headers_are_passed_to_subrequests(self):
        self.request.headers['Authorization'] = 'Basic ertyfghjkl'
        self.post({'requests': [{'path': '/'}]})
        subrequest, = self.request.invoke_subrequest.call_args[0]
        self.assertIn('Basic', subrequest.headers['Authorization'])

    def test_subrequests_body_are_json_serialized(self):
        request = {'path': '/', 'body': {'json': 'payload'}}
        self.post({'requests': [request]})
        wanted = {"json": "payload"}
        subrequest, = self.request.invoke_subrequest.call_args[0]
        self.assertEqual(subrequest.body.decode('utf8'),
                         json.dumps(wanted))

    def test_subrequests_body_have_json_content_type(self):
        self.request.headers['Content-Type'] = 'text/xml'
        request = {'path': '/', 'body': {'json': 'payload'}}
        self.post({'requests': [request]})
        subrequest, = self.request.invoke_subrequest.call_args[0]
        self.assertIn('application/json',
                      subrequest.headers['Content-Type'])

    def test_subrequests_body_have_utf8_charset(self):
        request = {'path': '/', 'body': {'json': "😂"}}
        self.post({'requests': [request]})
        subrequest, = self.request.invoke_subrequest.call_args[0]
        self.assertIn('charset=utf-8', subrequest.headers['Content-Type'])
        wanted = {"json": "😂"}
        self.assertEqual(subrequest.body.decode('utf8'),
                         json.dumps(wanted))

    def test_subrequests_paths_are_url_encoded(self):
        request = {'path': '/test?param=©'}
        self.post({'requests': [request]})
        subrequest, = self.request.invoke_subrequest.call_args[0]
        self.assertEqual(subrequest.path, '/v0/test')
        self.assertEqual(subrequest.GET['param'], '©')

    def test_subrequests_responses_paths_are_url_decoded(self):
        request = {'path': '/test?param=©'}
        resp = self.post({'requests': [request]})
        path = resp['responses'][0]['path']
        self.assertEqual(path, '/v0/test')

    def test_response_body_is_string_if_remote_response_is_not_json(self):
        response = Response(body='Internal Error')
        self.request.invoke_subrequest.return_value = response
        request = {'path': '/test'}
        resp = self.post({'requests': [request]})
        body = resp['responses'][0]['body'].decode('utf-8')
        self.assertEqual(body, 'Internal Error')

    def test_number_of_requests_is_not_limited_when_settings_set_to_none(self):
        self.request.registry.settings['batch_max_requests'] = None
        requests = {}
        for i in range(30):
            requests.setdefault('requests', []).append({'path': '/'})
            self.post(requests)

    def test_number_of_requests_is_limited_to_25_by_default(self):
        requests = {}
        for i in range(26):
            requests.setdefault('requests', []).append({'path': '/'})
        result = self.post(requests)

        self.assertEqual(self.request.errors[0]['description'],
                         'Number of requests is limited to 25')
        self.assertIsNone(result)  # rest of view not executed

    def test_return_400_if_number_of_requests_is_greater_than_settings(self):
        self.request.registry.settings['batch_max_requests'] = 22

        requests = {}
        for i in range(23):
            requests.setdefault('requests', []).append({'path': '/'})
        result = self.post(requests)

        self.assertEqual(self.request.errors[0]['description'],
                         'Number of requests is limited to 22')
        self.assertIsNone(result)  # rest of view not executed



import colander

from cornice.validators import colander_validator
from pyramid import httpexceptions
from pyramid.security import NO_PERMISSION_REQUIRED

from kinto.core import errors
from kinto.core import logger
from kinto.core import Service
from kinto.core.errors import ErrorSchema
from kinto.core.utils import merge_dicts, build_request, build_response


valid_http_method = colander.OneOf(('GET', 'HEAD', 'DELETE', 'TRACE',
                                    'POST', 'PUT', 'PATCH'))


def string_values(node, cstruct):
    """Validate that a ``colander.Mapping`` only has strings in its values.

    .. warning::

        Should be associated to a ``colander.Mapping`` schema node.
    """
    are_strings = [isinstance(v, str) for v in cstruct.values()]
    if not all(are_strings):
        error_msg = '{} contains non string value'.format(cstruct)
        raise colander.Invalid(node, error_msg)


class BatchRequestSchema(colander.MappingSchema):
    method = colander.SchemaNode(colander.String(),
                                 validator=valid_http_method,
                                 missing=colander.drop)
    path = colander.SchemaNode(colander.String(),
                               validator=colander.Regex('^/'))
    headers = colander.SchemaNode(colander.Mapping(unknown='preserve'),
                                  validator=string_values,
                                  missing=colander.drop)
    body = colander.SchemaNode(colander.Mapping(unknown='preserve'),
                               missing=colander.drop)


class BatchPayloadSchema(colander.MappingSchema):
    defaults = BatchRequestSchema(missing=colander.drop).clone()
    requests = colander.SchemaNode(colander.Sequence(),
                                   BatchRequestSchema())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # On defaults, path is not mandatory.
        self.get('defaults').get('path').missing = colander.drop

    def deserialize(self, cstruct=colander.null):
        """Preprocess received data to carefully merge defaults.
        """
        if cstruct is not colander.null:
            defaults = cstruct.get('defaults')
            requests = cstruct.get('requests')
            if isinstance(defaults, dict) and isinstance(requests, list):
                for request in requests:
                    if isinstance(request, dict):
                        merge_dicts(request, defaults)
        return super().deserialize(cstruct)


class BatchRequest(colander.MappingSchema):
    body = BatchPayloadSchema()


class BatchResponseSchema(colander.MappingSchema):
    status = colander.SchemaNode(colander.Integer())
    path = colander.SchemaNode(colander.String())
    headers = colander.SchemaNode(colander.Mapping(unknown='preserve'),
                                  validator=string_values,
                                  missing=colander.drop)
    body = colander.SchemaNode(colander.Mapping(unknown='preserve'),
                               missing=colander.drop)


class BatchResponseBodySchema(colander.MappingSchema):
    responses = colander.SequenceSchema(BatchResponseSchema(missing=colander.drop))


class BatchResponse(colander.MappingSchema):
    body = BatchResponseBodySchema()


class ErrorResponseSchema(colander.MappingSchema):
    body = ErrorSchema()


batch_responses = {
    '200': BatchResponse(description='Return a list of operation responses.'),
    '400': ErrorResponseSchema(description='The request was badly formatted.'),
    'default': ErrorResponseSchema(description='an unknown error occurred.')
}

batch = Service(name="batch", path='/batch',
                description="Batch operations")


@batch.post(schema=BatchRequest,
            validators=(colander_validator,),
            permission=NO_PERMISSION_REQUIRED,
            tags=['Batch'], operation_id='batch',
            response_schemas=batch_responses)
def post_batch(request):
    requests = request.validated['body']['requests']
    batch_size = len(requests)

    limit = request.registry.settings['batch_max_requests']
    if limit and len(requests) > int(limit):
        error_msg = 'Number of requests is limited to {}'.format(limit)
        request.errors.add('body', 'requests', error_msg)
        return

    if any([batch.path in req['path'] for req in requests]):
        error_msg = 'Recursive call on {} endpoint is forbidden.'.format(batch.path)
        request.errors.add('body', 'requests', error_msg)
        return

    responses = []

    sublogger = logger.new()

    for subrequest_spec in requests:
        subrequest = build_request(request, subrequest_spec)

        sublogger.bind(path=subrequest.path,
                       method=subrequest.method)
        try:
            # Invoke subrequest without individual transaction.
            resp, subrequest = request.follow_subrequest(subrequest,
                                                         use_tweens=False)
        except httpexceptions.HTTPException as e:
            if e.content_type == 'application/json':
                resp = e
            else:
                # JSONify raw Pyramid errors.
                resp = errors.http_error(e)

        sublogger.bind(code=resp.status_code)
        sublogger.info('subrequest.summary')

        dict_resp = build_response(resp, subrequest)
        responses.append(dict_resp)

    # Rebing batch request for summary
    logger.bind(path=batch.path,
                method=request.method,
                batch_size=batch_size,
                agent=request.headers.get('User-Agent'),)

    return {
        'responses': responses
    }



import mock
import time
import threading
import unittest
from uuid import uuid4

from pyramid import testing
from pyramid import httpexceptions

from kinto.core.storage.exceptions import BackendError
from kinto.core.utils import sqlalchemy
from kinto.core import events
from kinto.core.testing import skip_if_no_postgresql

from .support import BaseWebTest, USER_PRINCIPAL


class PostgreSQLTest(BaseWebTest):
    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        if sqlalchemy is not None:
            from .test_storage import PostgreSQLStorageTest
            from .test_cache import PostgreSQLCacheTest
            from .test_permission import PostgreSQLPermissionTest
            settings.update(**PostgreSQLStorageTest.settings)
            settings.update(**PostgreSQLCacheTest.settings)
            settings.update(**PostgreSQLPermissionTest.settings)
            settings.pop('storage_poolclass', None)
            settings.pop('cache_poolclass', None)
            settings.pop('permission_poolclass', None)
        return settings

    def run_failing_batch(self):
        patch = mock.patch.object(
            self.storage,
            'delete_all',
            side_effect=BackendError('boom'))
        self.addCleanup(patch.stop)
        patch.start()
        request_create = {
            'method': 'POST',
            'path': '/mushrooms',
            'body': {'data': {'name': 'Amanite'}}
        }
        request_delete = {
            'method': 'DELETE',
            'path': '/mushrooms'
        }
        body = {'requests': [request_create, request_create, request_delete]}
        self.app.post_json('/batch', body, headers=self.headers, status=503)

    def run_failing_post(self):
        patch = mock.patch.object(
            self.permission,
            'add_principal_to_ace',
            side_effect=BackendError('boom'))
        self.addCleanup(patch.stop)
        patch.start()
        self.app.post_json('/psilos',
                           {'data': {'name': 'Amanite'}},
                           headers=self.headers,
                           status=503)


@skip_if_no_postgresql
class TransactionTest(PostgreSQLTest, unittest.TestCase):

    def test_heartbeat_releases_transaction_lock(self):
        for i in range(4):
            # 4 calls because we have 3 backends
            # See bug Kinto/kinto#804
            self.app.get('/__heartbeat__')

    def test_storage_operations_are_committed_on_success(self):
        request_create = {
            'method': 'POST',
            'path': '/mushrooms',
            'body': {'data': {'name': 'Trompette de la mort'}}
        }
        body = {'requests': [request_create, request_create, request_create]}
        self.app.post_json('/batch', body, headers=self.headers)
        resp = self.app.get('/mushrooms', headers=self.headers)
        self.assertEqual(len(resp.json['data']), 3)

    def test_transaction_isolation_within_request(self):
        request_create = {
            'method': 'POST',
            'path': '/mushrooms',
            'body': {'data': {'name': 'Vesse de loup'}}
        }
        request_get = {
            'method': 'GET',
            'path': '/mushrooms',
        }
        body = {'requests': [request_create, request_get]}
        resp = self.app.post_json('/batch', body, headers=self.headers)
        self.assertEqual(resp.json['responses'][1]['body']['data'][0]['name'],
                         'Vesse de loup')

    def test_modifications_are_rolled_back_on_error(self):
        self.run_failing_batch()

        resp = self.app.get('/mushrooms', headers=self.headers)
        self.assertEqual(len(resp.json['data']), 0)

    def test_modifications_are_not_rolled_back_on_4XX_error(self):
        request_create = {
            'method': 'POST',
            'path': '/mushrooms',
            'body': {'data': {'name': 'Vesse de loup'}}
        }
        request_get = {
            'method': 'GET',
            'path': '/this-is-an-unknown-url',
        }
        body = {'requests': [request_create, request_get]}
        resp = self.app.post_json('/batch', body, headers=self.headers)
        resp = self.app.get('/mushrooms', headers=self.headers)
        self.assertEqual(len(resp.json['data']), 1)

    def test_modifications_are_rolled_back_on_error_accross_backends(self):
        self.run_failing_post()

        resp = self.app.get('/psilos', headers=self.headers)
        self.assertEqual(len(resp.json['data']), 0)


class IntegrityConstraintTest(PostgreSQLTest, unittest.TestCase):
    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        if sqlalchemy is not None:
            settings.pop('storage_poolclass', None)  # Use real pool.
        return settings

    def test_concurrent_transactions_do_not_fail(self):
        # This test originally intended to reproduce integrity errors and check
        # that a 409 was obtained. But since every errors that could be reproduced
        # could be also be fixed, this test just asserts that API responses
        # are consistent. # See Kinto/kinto#1125.

        # Make requests slow.
        patch = mock.patch('kinto.core.resource.UserResource.postprocess',
                           lambda s, r, a='read', old=None: time.sleep(0.2) or {})
        patch.start()
        self.addCleanup(patch.stop)

        # Same object created in two concurrent requests.
        body = {'data': {'id': str(uuid4())}}
        results = set()

        def create_object():
            r = self.app.post_json('/psilos', body, headers=self.headers,
                                   status=(201, 200, 409))
            results.add(r.status_code)

        thread1 = threading.Thread(target=create_object)
        thread2 = threading.Thread(target=create_object)
        thread1.start()
        time.sleep(0.1)
        thread2.start()
        thread1.join()
        thread2.join()
        self.assertTrue({201, 200, 409} >= results)


@skip_if_no_postgresql
class TransactionCacheTest(PostgreSQLTest, unittest.TestCase):
    def setUp(self):
        def cache_and_fails(this, *args, **kwargs):
            self.cache.set('test-cache', 'a value', ttl=100)
            raise BackendError('boom')

        patch = mock.patch.object(
            self.permission,
            'add_principal_to_ace',
            wraps=cache_and_fails)
        self.addCleanup(patch.stop)
        patch.start()

    def test_cache_backend_operations_are_always_committed(self):
        self.app.post_json('/psilos',
                           {'data': {'name': 'Amanite'}},
                           headers=self.headers,
                           status=503)

        # Storage was rolled back.
        resp = self.app.get('/psilos', headers=self.headers)
        self.assertEqual(len(resp.json['data']), 0)

        # Cache was committed.
        self.assertEqual(self.cache.get('test-cache'), 'a value')


@skip_if_no_postgresql
class TransactionEventsTest(PostgreSQLTest, unittest.TestCase):
    def make_app_with_subscribers(self, subscribers):
        settings = self.get_app_settings({})
        config = testing.setUp(settings=settings)
        for event, subscriber in subscribers:
            config.add_subscriber(subscriber, event)
        config.commit()
        return self.make_app(config=config)

    def send_batch_create(self, app, **kwargs):
        body = {
            "defaults": {
                "method": "POST",
                "body": {'data': {'name': 'Vesse de loup'}},
            },
            "requests": [
                {"path": '/mushrooms'},
                {"path": '/mushrooms'}
            ]
        }
        return app.post_json("/batch", body, headers=self.headers, **kwargs)

    def test_resourcechanged_is_executed_within_transaction(self):
        def store_record(event):
            storage = event.request.registry.storage
            extra_record = {"id": "3.14", "z": 42}
            storage.create("mushroom", USER_PRINCIPAL, extra_record)

        app = self.make_app_with_subscribers([(events.ResourceChanged,
                                               store_record)])
        self.send_batch_create(app)
        resp = app.get('/mushrooms', headers=self.headers)
        self.assertEqual(len(resp.json['data']), 2 + 1)

    def test_resourcechanged_is_rolledback_with_transaction(self):
        def store_record(event):
            storage = event.request.registry.storage
            extra_record = {"id": "3.14", "z": 42}
            storage.create("mushroom", USER_PRINCIPAL, extra_record)

        app = self.make_app_with_subscribers([(events.ResourceChanged,
                                               store_record)])
        # We want to patch the following method so that it raises an exception,
        # to make sure the rollback is happening correctly.
        to_patch = 'transaction._manager.ThreadTransactionManager.commit'
        with mock.patch(to_patch) as mocked:
            mocked.side_effect = ValueError
            self.send_batch_create(app, status=500)
        resp = app.get('/mushrooms', headers=self.headers)
        self.assertEqual(len(resp.json['data']), 0)

    def test_resourcechanged_can_rollback_whole_request(self):
        def store_record(event):
            raise httpexceptions.HTTPInsufficientStorage()

        app = self.make_app_with_subscribers([(events.ResourceChanged,
                                               store_record)])
        self.send_batch_create(app, status=507)
        resp = app.get('/mushrooms', headers=self.headers)
        self.assertEqual(len(resp.json['data']), 0)

    def test_afterresourcechanged_cannot_rollback_whole_request(self):
        def store_record(event):
            raise httpexceptions.HTTPInsufficientStorage()

        app = self.make_app_with_subscribers([(events.AfterResourceChanged,
                                               store_record)])
        self.send_batch_create(app, status=200)
        resp = app.get('/mushrooms', headers=self.headers)
        self.assertEqual(len(resp.json['data']), 2)


@skip_if_no_postgresql
class WithoutTransactionTest(PostgreSQLTest, unittest.TestCase):

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['transaction_per_request'] = False
        return settings

    def test_modifications_are_not_rolled_back_on_error(self):
        self.run_failing_batch()

        resp = self.app.get('/mushrooms', headers=self.headers)
        self.assertEqual(len(resp.json['data']), 2)

    def test_modifications_are_not_rolled_back_on_error_accross_backends(self):
        self.run_failing_post()

        resp = self.app.get('/psilos', headers=self.headers)
        self.assertEqual(len(resp.json['data']), 1)



from . import BaseTest


class CacheExpires(BaseTest):
    setting = 'test_cache_expires_seconds'

    def get_context(self):
        context = super().get_context()
        context.resource_name = 'test'
        return context

    def get_request(self):
        request = super().get_request()
        request.prefixed_userid = None  # Anonymous.
        return request

    def test_no_cache_expires_by_default(self):
        settings = self.resource.request.registry.settings
        settings.pop(self.setting, None)
        self.resource.collection_get()
        self.assertFalse(self.last_response.cache_expires.called)

    def test_cache_expires_using_setting_with_resource_name(self):
        settings = self.resource.request.registry.settings
        settings[self.setting] = 3
        self.resource.collection_get()
        self.last_response.cache_expires.assert_called_with(seconds=3)

    def test_cache_expires_is_also_on_record_get(self):
        stored = self.model.create_record({})
        self.resource.record_id = stored['id']
        settings = self.resource.request.registry.settings
        settings[self.setting] = 3
        self.resource.get()
        self.last_response.cache_expires.assert_called_with(seconds=3)

    def test_cache_expires_even_if_zero(self):
        settings = self.resource.request.registry.settings
        settings[self.setting] = 0
        self.resource.collection_get()
        self.assertTrue(self.last_response.cache_control.no_cache)



import unittest
from kinto.core.errors import ERRORS
from kinto.core.testing import FormattedErrorMixin

from .support import (BaseWebTest, MINIMALIST_BUCKET,
                      MINIMALIST_GROUP)


class GroupViewTest(FormattedErrorMixin, BaseWebTest, unittest.TestCase):

    collection_url = '/buckets/beers/groups'
    record_url = '/buckets/beers/groups/moderators'

    def setUp(self):
        super().setUp()
        self.app.put_json('/buckets/beers', MINIMALIST_BUCKET,
                          headers=self.headers)
        resp = self.app.put_json(self.record_url,
                                 MINIMALIST_GROUP,
                                 headers=self.headers)
        self.record = resp.json['data']

    def test_collection_endpoint_lists_them_all(self):
        resp = self.app.get(self.collection_url, headers=self.headers)
        records = resp.json['data']
        self.assertEqual(len(records), 1)
        self.assertEqual(records[0]['members'], ['fxa:user'])

    def test_groups_can_be_posted_without_id(self):
        resp = self.app.post_json(self.collection_url,
                                  MINIMALIST_GROUP,
                                  headers=self.headers,
                                  status=201)
        self.assertIn('id', resp.json['data'])
        self.assertEqual(resp.json['data']['members'], ['fxa:user'])

    def test_groups_can_be_put_with_simple_name(self):
        self.assertEqual(self.record['id'], 'moderators')

    def test_groups_name_should_be_simple(self):
        self.app.put_json('/buckets/beers/groups/__moderator__',
                          MINIMALIST_GROUP,
                          headers=self.headers,
                          status=400)

    def test_groups_can_have_arbitrary_attributes(self):
        mailinglist = "kinto@mozilla.com"
        group = {**MINIMALIST_GROUP, 'data': {**MINIMALIST_GROUP['data'],
                                              'mailinglist': mailinglist}}
        resp = self.app.put_json('/buckets/beers/groups/moderator',
                                 group,
                                 headers=self.headers)
        data = resp.json['data']
        self.assertIn('mailinglist', data)
        self.assertEqual(data['mailinglist'], mailinglist)

    def test_groups_can_be_filtered_by_arbitrary_attribute(self):
        group = {**MINIMALIST_GROUP, 'data': {**MINIMALIST_GROUP['data'], 'size': 3}}
        self.app.put_json('/buckets/beers/groups/moderator',
                          group,
                          headers=self.headers)
        resp = self.app.get('/buckets/beers/groups?min_size=2',
                            headers=self.headers)
        data = resp.json['data']
        self.assertEqual(len(data), 1)

    def test_groups_should_reject_unaccepted_request_content_type(self):
        headers = {**self.headers, 'Content-Type': 'text/plain'}
        self.app.put('/buckets/beers/groups/moderator',
                     MINIMALIST_GROUP,
                     headers=headers,
                     status=415)

    def test_unknown_bucket_raises_403(self):
        other_bucket = self.collection_url.replace('beers', 'sodas')
        self.app.get(other_bucket, headers=self.headers, status=403)

    def test_groups_are_isolated_by_bucket(self):
        other_bucket = self.record_url.replace('beers', 'sodas')
        self.app.put_json('/buckets/sodas',
                          MINIMALIST_BUCKET,
                          headers=self.headers)
        self.app.get(other_bucket, headers=self.headers, status=404)

    def test_wrong_create_permissions_cannot_be_added_on_groups(self):
        group = {**MINIMALIST_GROUP, 'permissions': {'group:create': ['fxa:user']}}
        self.app.put_json('/buckets/beers/groups/moderator',
                          group,
                          headers=self.headers,
                          status=400)

    def test_recreate_group_after_deletion_returns_a_201(self):
        self.app.put_json('/buckets/beers/groups/moderator',
                          MINIMALIST_GROUP,
                          headers=self.headers,
                          status=201)
        self.app.delete('/buckets/beers/groups/moderator',
                        headers=self.headers,
                        status=200)
        self.app.put_json('/buckets/beers/groups/moderator',
                          MINIMALIST_GROUP,
                          headers=self.headers,
                          status=201)

    def test_group_doesnt_accept_system_Everyone(self):
        group = {**MINIMALIST_GROUP, 'data': {'members': ['system.Everyone']}}
        response = self.app.put_json('/buckets/beers/groups/moderator',
                                     group,
                                     headers=self.headers,
                                     status=400)
        self.assertFormattedError(
            response, 400, ERRORS.INVALID_PARAMETERS,
            "Invalid parameters",
            "'system.Everyone' is not a valid user ID.")

    def test_group_doesnt_accept_groups_inside_groups(self):
        group = {**MINIMALIST_GROUP, 'data': {'members': ['/buckets/beers/groups/administrators']}}
        response = self.app.put_json('/buckets/beers/groups/moderator',
                                     group,
                                     headers=self.headers,
                                     status=400)
        self.assertFormattedError(
            response, 400, ERRORS.INVALID_PARAMETERS,
            "Invalid parameters",
            "'/buckets/beers/groups/administrators' is not a valid user ID.")


class GroupManagementTest(BaseWebTest, unittest.TestCase):

    group_url = '/buckets/beers/groups/moderators'

    def setUp(self):
        super().setUp()
        self.create_bucket('beers')

    def test_groups_can_be_deleted(self):
        self.create_group('beers', 'moderators')
        self.app.delete(self.group_url, headers=self.headers)
        self.app.get(self.group_url, headers=self.headers,
                     status=404)

    def test_unknown_group_raises_404(self):
        other_group = self.group_url.replace('moderators', 'blah')
        resp = self.app.get(other_group, headers=self.headers, status=404)
        self.assertEqual(resp.json['details']['id'], 'blah')
        self.assertEqual(resp.json['details']['resource_name'], 'group')

    def test_group_is_removed_from_users_principals_on_group_deletion(self):
        self.app.put_json(self.group_url, MINIMALIST_GROUP,
                          headers=self.headers, status=201)
        self.assertIn(self.group_url,
                      self.permission.get_user_principals('fxa:user'))
        self.app.delete(self.group_url, headers=self.headers, status=200)
        self.assertNotIn(self.group_url,
                         self.permission.get_user_principals('fxa:user'))

    def test_group_is_removed_from_users_principals_on_groups_deletion(self):
        self.create_group('beers', 'moderators', ['natim', 'fxa:me'])
        self.create_group('beers', 'reviewers', ['natim', 'alexis'])

        self.app.delete('/buckets/beers/groups', headers=self.headers,
                        status=200)

        self.assertEquals(self.permission.get_user_principals('fxa:me'), set())
        self.assertEquals(self.permission.get_user_principals('natim'), set())
        self.assertEquals(self.permission.get_user_principals('alexis'), set())

    def test_group_is_added_to_user_principals_when_added_to_members(self):
        self.create_group('beers', 'moderators', ['natim', 'mat'])

        self.app.get('/buckets/beers/groups', headers=self.headers, status=200)
        self.assertEquals(self.permission.get_user_principals('natim'),
                          {'/buckets/beers/groups/moderators'})
        self.assertEquals(self.permission.get_user_principals('mat'),
                          {'/buckets/beers/groups/moderators'})

    def test_group_is_added_to_user_principals_on_members_add_with_patch(self):
        self.create_group('beers', 'moderators', ['natim', 'mat'])
        group_url = '/buckets/beers/groups/moderators'
        group = {'data': {'members': ['natim', 'mat', 'alice']}}
        self.app.patch_json(group_url, group,
                            headers=self.headers, status=200)
        self.app.get('/buckets/beers/groups', headers=self.headers, status=200)
        self.assertEquals(self.permission.get_user_principals('natim'),
                          {group_url})
        self.assertEquals(self.permission.get_user_principals('mat'),
                          {group_url})
        self.assertEquals(self.permission.get_user_principals('alice'),
                          {group_url})

    def test_group_member_removal_updates_user_principals(self):
        self.create_group('beers', 'moderators', ['natim', 'mat'])
        group_url = '/buckets/beers/groups/moderators'
        group = {'data': {'members': ['mat']}}
        self.app.put_json(group_url, group,
                          headers=self.headers, status=200)
        self.app.get('/buckets/beers/groups', headers=self.headers, status=200)
        self.assertEquals(self.permission.get_user_principals('natim'), set())
        self.assertEquals(self.permission.get_user_principals('mat'),
                          {group_url})

    def test_group_with_authenticated_is_added_to_everbody(self):
        self.create_group('beers', 'reviewers', ['system.Authenticated'])
        self.assertEquals(self.permission.get_user_principals('natim'),
                          {'/buckets/beers/groups/reviewers'})
        self.assertEquals(self.permission.get_user_principals('mat'),
                          {'/buckets/beers/groups/reviewers'})

    def test_group_member_removal_updates_user_principals_with_patch(self):
        self.create_group('beers', 'moderators', ['natim', 'mat'])
        group_url = '/buckets/beers/groups/moderators'
        group = {'data': {'members': ['mat']}}
        self.app.patch_json(group_url, group, headers=self.headers, status=200)
        self.assertEquals(self.permission.get_user_principals('natim'), set())
        self.assertEquals(self.permission.get_user_principals('mat'),
                          {group_url})

    def test_groups_can_be_created_after_deletion(self):
        self.create_group('beers', 'moderators')
        group_url = '/buckets/beers/groups/moderators'
        self.app.delete(group_url, headers=self.headers)
        headers = {**self.headers, 'If-None-Match': '*'}
        self.app.put_json(group_url, MINIMALIST_GROUP,
                          headers=headers, status=201)

    def test_groups_data_is_optional_with_patch(self):
        valid = {'permissions': {'write': ['github:me']}}
        self.app.put_json(self.group_url, MINIMALIST_GROUP,
                          headers=self.headers)
        self.app.patch_json(self.group_url,
                            valid,
                            headers=self.headers)


class InvalidGroupTest(BaseWebTest, unittest.TestCase):

    group_url = '/buckets/beers/groups/moderators'

    def setUp(self):
        super().setUp()
        self.create_bucket('beers')

    def test_groups_data_is_required_with_put(self):
        invalid = {'permissions': {'write': ['github:me']}}
        resp = self.app.put_json(self.group_url,
                                 invalid,
                                 headers=self.headers,
                                 status=400)
        self.assertEqual(resp.json['message'], "data in body: Required")

    def test_groups_must_have_members_attribute(self):
        invalid = {'data': {'alias': 'admins'}}
        resp = self.app.put_json(self.group_url,
                                 invalid,
                                 headers=self.headers,
                                 status=400)
        self.assertIn('data.members in body', resp.json['message'])



from bravado_core.response import validate_response

from kinto.core import testing
from .support import SwaggerTest
from ..support import (MINIMALIST_BUCKET, MINIMALIST_GROUP,
                       MINIMALIST_COLLECTION, MINIMALIST_RECORD)


class SwaggerObjectErrorResponsesTest(SwaggerTest):

    def setUp(self):
        super().setUp()

        self.bucket = self.app.put_json('/buckets/b1',
                                        MINIMALIST_BUCKET,
                                        headers=self.headers).json

        self.group = self.app.put_json('/buckets/b1/groups/g1',
                                       MINIMALIST_GROUP,
                                       headers=self.headers).json

        self.collection = self.app.put_json('/buckets/b1/collections/c1',
                                            MINIMALIST_COLLECTION,
                                            headers=self.headers).json

        self.record = self.app.put_json('/buckets/b1/collections/c1/records/r1',
                                        MINIMALIST_RECORD,
                                        headers=self.headers).json

    def test_object_get_304(self):
        headers = {**self.headers,
                   'If-None-Match': '"{}"'.format(self.bucket['data']['last_modified'])}
        response = self.app.get('/buckets/b1',
                                headers=headers, status=304)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].get_bucket
        schema = self.spec.deref(op.op_spec['responses']['304'])
        validate_response(schema, op, response)

    def test_object_get_400(self):
        headers = {**self.headers, 'If-None-Match': 'aaa'}
        response = self.app.get('/buckets/b1',
                                headers=headers, status=400)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].get_bucket
        schema = self.spec.deref(op.op_spec['responses']['400'])
        validate_response(schema, op, response)

    def test_object_get_401(self):
        response = self.app.get('/buckets/b1', status=401)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].get_bucket
        schema = self.spec.deref(op.op_spec['responses']['401'])
        validate_response(schema, op, response)

    def test_object_get_403(self):
        headers = {**self.headers, **testing.get_user_headers('aaa')}
        response = self.app.get('/buckets/b1',
                                headers=headers, status=403)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].get_bucket
        schema = self.spec.deref(op.op_spec['responses']['403'])
        validate_response(schema, op, response)

    def test_object_get_404(self):
        response = self.app.get('/buckets/b1/collections/col',
                                headers=self.headers, status=404)
        response = self.cast_bravado_response(response)
        op = self.resources['Collections'].get_collection
        schema = self.spec.deref(op.op_spec['responses']['404'])
        validate_response(schema, op, response)

    def test_object_get_406(self):
        headers = {**self.headers, 'Accept': 'text/html'}
        response = self.app.get('/buckets/b1',
                                headers=headers, status=406)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].get_bucket
        schema = self.spec.deref(op.op_spec['responses']['406'])
        validate_response(schema, op, response)

    def test_object_get_412(self):
        headers = {**self.headers,
                   'If-Match': '"{}"'.format(self.bucket['data']['last_modified']-1)}
        response = self.app.get('/buckets/b1',
                                headers=headers, status=412)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].get_bucket
        schema = self.spec.deref(op.op_spec['responses']['412'])
        validate_response(schema, op, response)

    def test_object_put_400(self):
        response = self.app.put_json('/buckets/b1', [],
                                     headers=self.headers, status=400)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].update_bucket
        schema = self.spec.deref(op.op_spec['responses']['400'])
        validate_response(schema, op, response)

    def test_object_put_401(self):
        response = self.app.put_json('/buckets/b1', MINIMALIST_BUCKET, status=401)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].update_bucket
        schema = self.spec.deref(op.op_spec['responses']['401'])
        validate_response(schema, op, response)

    def test_object_put_403(self):
        headers = {**self.headers, **testing.get_user_headers('aaa')}
        response = self.app.put_json('/buckets/b1', MINIMALIST_BUCKET,
                                     headers=headers, status=403)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].update_bucket
        schema = self.spec.deref(op.op_spec['responses']['403'])
        validate_response(schema, op, response)

    def test_object_put_406(self):
        headers = {**self.headers, 'Accept': 'text/html'}
        response = self.app.put_json('/buckets/b1', MINIMALIST_BUCKET,
                                     headers=headers, status=406)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].update_bucket
        schema = self.spec.deref(op.op_spec['responses']['406'])
        validate_response(schema, op, response)

    def test_object_put_412(self):
        headers = {**self.headers,
                   'If-Match': '"{}"'.format(self.bucket['data']['last_modified']-1)}
        response = self.app.put_json('/buckets/b1', MINIMALIST_BUCKET,
                                     headers=headers, status=412)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].update_bucket
        schema = self.spec.deref(op.op_spec['responses']['412'])
        validate_response(schema, op, response)

    def test_object_put_415(self):
        headers = {**self.headers, 'Content-Type': 'text/html'}
        response = self.app.put_json('/buckets/b1', MINIMALIST_BUCKET,
                                     headers=headers, status=415)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].update_bucket
        schema = self.spec.deref(op.op_spec['responses']['415'])
        validate_response(schema, op, response)

    def test_object_patch_400(self):
        response = self.app.patch_json('/buckets/b1', [],
                                       headers=self.headers, status=400)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].patch_bucket
        schema = self.spec.deref(op.op_spec['responses']['400'])
        validate_response(schema, op, response)

    def test_object_patch_401(self):
        response = self.app.patch_json('/buckets/b1', MINIMALIST_BUCKET, status=401)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].patch_bucket
        schema = self.spec.deref(op.op_spec['responses']['401'])
        validate_response(schema, op, response)

    def test_object_patch_403(self):
        headers = {**self.headers, **testing.get_user_headers('aaa')}
        response = self.app.patch_json('/buckets/b1', MINIMALIST_BUCKET,
                                       headers=headers, status=403)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].patch_bucket
        schema = self.spec.deref(op.op_spec['responses']['403'])
        validate_response(schema, op, response)

    def test_object_patch_404(self):
        response = self.app.patch_json('/buckets/b1/collections/col',
                                       MINIMALIST_COLLECTION,
                                       headers=self.headers, status=404)
        response = self.cast_bravado_response(response)
        op = self.resources['Collections'].patch_collection
        schema = self.spec.deref(op.op_spec['responses']['404'])
        validate_response(schema, op, response)

    def test_object_patch_406(self):
        headers = {**self.headers, 'Accept': 'text/html'}
        response = self.app.patch_json('/buckets/b1', MINIMALIST_BUCKET,
                                       headers=headers, status=406)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].patch_bucket
        schema = self.spec.deref(op.op_spec['responses']['406'])
        validate_response(schema, op, response)

    def test_object_patch_412(self):
        headers = {**self.headers,
                   'If-Match': '"{}"'.format(self.bucket['data']['last_modified']-1)}
        response = self.app.patch_json('/buckets/b1', MINIMALIST_BUCKET,
                                       headers=headers, status=412)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].patch_bucket
        schema = self.spec.deref(op.op_spec['responses']['412'])
        validate_response(schema, op, response)

    def test_object_patch_415(self):
        headers = {**self.headers, 'Content-Type': 'text/html'}
        response = self.app.patch_json('/buckets/b1', MINIMALIST_BUCKET,
                                       headers=headers, status=415)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].patch_bucket
        schema = self.spec.deref(op.op_spec['responses']['415'])
        validate_response(schema, op, response)

    def test_object_delete_400(self):
        headers = {**self.headers, 'If-Match': 'aaa'}
        response = self.app.delete('/buckets/b1',
                                   headers=headers, status=400)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].delete_bucket
        schema = self.spec.deref(op.op_spec['responses']['400'])
        validate_response(schema, op, response)

    def test_object_delete_401(self):
        response = self.app.delete('/buckets/b1', status=401)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].delete_bucket
        schema = self.spec.deref(op.op_spec['responses']['401'])
        validate_response(schema, op, response)

    def test_object_delete_403(self):
        headers = {**self.headers, **testing.get_user_headers('aaa')}
        response = self.app.delete('/buckets/b1',
                                   headers=headers, status=403)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].delete_bucket
        schema = self.spec.deref(op.op_spec['responses']['403'])
        validate_response(schema, op, response)

    def test_object_delete_404(self):
        response = self.app.delete('/buckets/b1/collections/col',
                                   headers=self.headers, status=404)
        response = self.cast_bravado_response(response)
        op = self.resources['Collections'].delete_collection
        schema = self.spec.deref(op.op_spec['responses']['404'])
        validate_response(schema, op, response)

    def test_object_delete_406(self):
        headers = {**self.headers, 'Accept': 'text/html'}
        response = self.app.delete('/buckets/b1',
                                   headers=headers, status=406)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].delete_bucket
        schema = self.spec.deref(op.op_spec['responses']['406'])
        validate_response(schema, op, response)

    def test_object_delete_412(self):
        headers = {**self.headers,
                   'If-Match': '"{}"'.format(self.bucket['data']['last_modified']-1)}
        response = self.app.delete('/buckets/b1',
                                   headers=headers, status=412)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].delete_bucket
        schema = self.spec.deref(op.op_spec['responses']['412'])
        validate_response(schema, op, response)

    def test_list_get_304(self):
        headers = {**self.headers,
                   'If-None-Match': '"{}"'.format(self.bucket['data']['last_modified'])}
        response = self.app.get('/buckets',
                                headers=headers, status=304)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].get_buckets
        schema = self.spec.deref(op.op_spec['responses']['304'])
        validate_response(schema, op, response)

    def test_list_get_400(self):
        response = self.app.get('/buckets?_since=aaa',
                                headers=self.headers, status=400)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].get_buckets
        schema = self.spec.deref(op.op_spec['responses']['400'])
        validate_response(schema, op, response)

    def test_list_get_401(self):
        response = self.app.get('/buckets', status=401)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].get_buckets
        schema = self.spec.deref(op.op_spec['responses']['401'])
        validate_response(schema, op, response)

    def test_list_get_403(self):
        headers = {**self.headers, **testing.get_user_headers('aaa')}
        response = self.app.get('/buckets/b1/collections',
                                headers=headers, status=403)
        response = self.cast_bravado_response(response)
        op = self.resources['Collections'].get_collections
        schema = self.spec.deref(op.op_spec['responses']['403'])
        validate_response(schema, op, response)

    def test_list_get_406(self):
        headers = {**self.headers, 'Accept': 'text/html'}
        response = self.app.get('/buckets',
                                headers=headers, status=406)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].get_buckets
        schema = self.spec.deref(op.op_spec['responses']['406'])
        validate_response(schema, op, response)

    def test_list_get_412(self):
        headers = {**self.headers,
                   'If-Match': '"{}"'.format(self.bucket['data']['last_modified']-1)}
        response = self.app.get('/buckets/b1',
                                headers=headers, status=412)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].get_buckets
        schema = self.spec.deref(op.op_spec['responses']['412'])
        validate_response(schema, op, response)

    def test_list_delete_401(self):
        response = self.app.delete('/buckets', status=401)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].delete_buckets
        schema = self.spec.deref(op.op_spec['responses']['401'])
        validate_response(schema, op, response)

    def test_lidt_delete_403(self):
        headers = {**self.headers, **testing.get_user_headers('aaa')}
        response = self.app.delete('/buckets/b1/collections',
                                   headers=headers, status=403)
        response = self.cast_bravado_response(response)
        op = self.resources['Collections'].delete_collections
        schema = self.spec.deref(op.op_spec['responses']['403'])
        validate_response(schema, op, response)

    def test_list_delete_405(self):
        # XXX
        pass

    def test_list_delete_406(self):
        headers = {**self.headers, 'Accept': 'text/html'}
        response = self.app.delete('/buckets',
                                   headers=headers, status=406)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].delete_buckets
        schema = self.spec.deref(op.op_spec['responses']['406'])
        validate_response(schema, op, response)

    def test_list_delete_412(self):
        headers = {**self.headers,
                   'If-Match': '"{}"'.format(self.bucket['data']['last_modified']-1)}
        response = self.app.delete('/buckets',
                                   headers=headers, status=412)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].delete_buckets
        schema = self.spec.deref(op.op_spec['responses']['412'])
        validate_response(schema, op, response)



from .support import (SwaggerTest, MINIMALIST_BUCKET, MINIMALIST_GROUP,
                      MINIMALIST_COLLECTION, MINIMALIST_RECORD)


class SwaggerResourcesTest(SwaggerTest):

    allowed_failures = ['version']

    def test_resource_utilities(self):
        resource = self.resources['Utilities']
        for op_id, op in resource.operations.items():
                if op_id not in self.allowed_failures:
                    self.validate_request_call(op)

    def test_resource_batch(self):
        resource = self.resources['Batch']
        for op in resource.operations.values():
            requests = [{'path': '/v1/buckets'}]
            defaults = {'method': 'POST'}
            self.request.json = lambda: dict(
                requests=requests,
                defaults=defaults
            )
            self.validate_request_call(op)

    def test_resource_buckets(self):
        resource = self.resources['Buckets']

        for op in resource.operations.values():
            self.app.put_json('/buckets/b1',
                              MINIMALIST_BUCKET, headers=self.headers)
            self.request.path = {
                'id': 'b1',
            }
            bucket = {**MINIMALIST_BUCKET, 'data': {'foo': 'bar'}}
            self.request.json = lambda: bucket
            self.validate_request_call(op)

    def test_resource_groups(self):
        resource = self.resources['Groups']

        for op in resource.operations.values():
            self.app.put_json('/buckets/b1/groups/g1',
                              MINIMALIST_GROUP, headers=self.headers)
            self.request.path = {
                'bucket_id': 'b1',
                'id': 'g1',
            }
            self.request.json = lambda: MINIMALIST_GROUP
            self.validate_request_call(op)

    def test_resource_collections(self):
        resource = self.resources['Collections']

        for op in resource.operations.values():
            self.app.put_json('/buckets/b1/collections/c1',
                              MINIMALIST_COLLECTION, headers=self.headers)
            self.request.path = {
                'bucket_id': 'b1',
                'id': 'c1',
            }
            collection = {**MINIMALIST_COLLECTION, 'data': {'foo': 'bar'}}
            self.request.json = lambda: collection
            self.validate_request_call(op)

    def test_resource_records(self):
        resource = self.resources['Records']

        for op in resource.operations.values():
            self.app.put_json('/buckets/b1/collections/c1/records/r1',
                              MINIMALIST_RECORD, headers=self.headers)
            self.request.path = {
                'bucket_id': 'b1',
                'collection_id': 'c1',
                'id': 'r1',
            }
            self.request.json = lambda: MINIMALIST_RECORD
            self.validate_request_call(op)



import json
import os

from pyramid.settings import aslist

from kinto.core.decorators import cache_forever

HERE = os.path.abspath(os.path.dirname(__file__))


# Configured home page
@cache_forever
def admin_home_view(request):
    settings = {
        "authMethods": aslist(request.registry.settings.get('multiauth.policies'))
    }
    globalSettings = "<script>window.globalSettings = {};</script>".format(json.dumps(settings))

    # Update the file built by react-scripts to load the globalSettings.
    with open(os.path.join(HERE, 'build/index.html')) as f:
        return f.read().replace('<script', globalSettings + '<script')



from pyramid import httpexceptions

from kinto.core.errors import ERRORS

from . import BaseTest


class NotModifiedTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.stored = self.model.create_record({})

        self.resource = self.resource_class(request=self.get_request(),
                                            context=self.get_context())
        self.resource.request.validated = {**self.validated}
        self.resource.collection_get()
        self.validated = self.resource.request.validated
        current = self.last_response.headers['ETag'][1:-1]
        self.validated['header']['If-None-Match'] = int(current)

    def test_collection_returns_200_if_change_meanwhile(self):
        self.validated['header']['If-None-Match'] = 42
        self.resource.collection_get()  # Not raising.

    def test_collection_returns_304_if_no_change_meanwhile(self):
        try:
            self.resource.collection_get()
        except httpexceptions.HTTPNotModified as e:
            error = e
        self.assertEqual(error.code, 304)
        self.assertIsNotNone(error.headers.get('ETag'))
        self.assertIsNotNone(error.headers.get('Last-Modified'))

    def test_single_record_returns_304_if_no_change_meanwhile(self):
        self.resource.record_id = self.stored['id']
        try:
            self.resource.get()
        except httpexceptions.HTTPNotModified as e:
            error = e
        self.assertEqual(error.code, 304)
        self.assertIsNotNone(error.headers.get('ETag'))
        self.assertIsNotNone(error.headers.get('Last-Modified'))

    def test_single_record_last_modified_is_returned(self):
        self.resource.timestamp = 0
        self.resource.record_id = self.stored['id']
        try:
            self.resource.get()
        except httpexceptions.HTTPNotModified as e:
            error = e
        self.assertNotIn('1970', error.headers['Last-Modified'])


class ModifiedMeanwhileTest(BaseTest):
    def setUp(self):
        super().setUp()

        # Create a record using model (will have an incremented last_modified)
        self.stored = self.model.create_record({})
        # Update the record we just created (will set last_modified with the server ETag)
        self.resource.record_id = self.stored['id']
        self.resource.put()
        # Update our record with last_modified provided by the server
        self.stored['last_modified'] = int(self.last_response.headers['ETag'][1:-1])

        self.validated = self.resource.request.validated
        self.current = int(self.last_response.headers['ETag'][1:-1])
        previous = self.current - 10
        self.validated['header']['If-Match'] = previous

    def test_preconditions_errors_are_json_formatted(self):
        try:
            self.resource.collection_get()
        except httpexceptions.HTTPPreconditionFailed as e:
            error = e
        self.assertEqual(error.json, {
            'errno': ERRORS.MODIFIED_MEANWHILE.value,
            'message': 'Resource was modified meanwhile',
            'code': 412,
            'error': 'Precondition Failed'})

    def test_collection_returns_412_if_changed_meanwhile(self):
        self.assertRaises(httpexceptions.HTTPPreconditionFailed,
                          self.resource.collection_get)

    def test_collection_returns_412_if_if_match_is_superior(self):
        self.validated['header']['If-Match'] = self.current + 10
        self.assertRaises(httpexceptions.HTTPPreconditionFailed,
                          self.resource.collection_get)

    def test_collection_returns_200_if_if_match_is_equal(self):
        self.validated['header']['If-Match'] = self.current
        self.resource.collection_get()

    def test_412_errors_on_collection_do_not_provide_existing_record(self):
        try:
            self.resource.collection_get()
        except httpexceptions.HTTPPreconditionFailed as e:
            error = e
        self.assertNotIn('existing', error.json.get('details', {}))

    def test_412_on_collection_has_last_modified_timestamp(self):
        try:
            self.resource.collection_get()
        except httpexceptions.HTTPPreconditionFailed as e:
            error = e
        self.assertIsNotNone(error.headers.get('ETag'))
        self.assertIsNotNone(error.headers.get('Last-Modified'))

    def test_412_errors_on_record_provide_existing_data(self):
        try:
            self.resource.put()
        except httpexceptions.HTTPPreconditionFailed as e:
            error = e
        print(error.json['details']['existing'])
        print(self.stored)
        self.assertDictEqual(error.json['details']['existing'],
                             self.stored)

    def test_single_record_returns_412_if_changed_meanwhile(self):
        self.assertRaises(httpexceptions.HTTPPreconditionFailed,
                          self.resource.get)

    def test_412_on_single_record_has_last_modified_timestamp(self):
        try:
            self.resource.get()
        except httpexceptions.HTTPPreconditionFailed as e:
            error = e
        self.assertIsNotNone(error.headers.get('ETag'))
        self.assertIsNotNone(error.headers.get('Last-Modified'))

    def test_create_returns_412_if_changed_meanwhile(self):
        self.validated['body'] = {'data': {'field': 'new'}}
        self.assertRaises(httpexceptions.HTTPPreconditionFailed,
                          self.resource.collection_post)

    def test_put_returns_412_if_changed_meanwhile(self):
        self.model.delete_record(self.stored)
        self.assertRaises(httpexceptions.HTTPPreconditionFailed,
                          self.resource.put)

    def test_put_returns_412_if_changed_and_none_match_present(self):
        self.validated['body'] = {'data': {'field': 'new'}}
        self.validated['header']['If-None-Match'] = 42
        self.assertRaises(httpexceptions.HTTPPreconditionFailed,
                          self.resource.put)

    def test_put_returns_last_modified_if_changed_meanwhile(self):
        self.resource.timestamp = 0
        self.resource.record_id = self.stored['id']
        try:
            self.resource.put()
        except httpexceptions.HTTPPreconditionFailed as e:
            error = e
        self.assertNotIn('1970', error.headers['Last-Modified'])

    def test_put_returns_412_if_deleted_meanwhile(self):
        self.assertRaises(httpexceptions.HTTPPreconditionFailed,
                          self.resource.put)

    def test_put_if_none_match_star_fails_if_record_exists(self):
        self.validated['header'].pop('If-Match')
        self.validated['header']['If-None-Match'] = '*'
        self.assertRaises(httpexceptions.HTTPPreconditionFailed,
                          self.resource.put)

    def test_get_if_none_match_star_fails_if_record_exists(self):
        self.validated['header'].pop('If-Match')
        self.validated['header']['If-None-Match'] = '*'
        self.assertRaises(httpexceptions.HTTPPreconditionFailed,
                          self.resource.get)

    def test_put_if_none_match_star_succeeds_if_record_does_not_exist(self):
        self.validated['header'].pop('If-Match')
        self.validated['header']['If-None-Match'] = '*'
        self.validated['body'] = {'data': {'field': 'new'}}
        self.resource.record_id = self.resource.model.id_generator()
        self.resource.put()  # not raising.

    def test_put_if_none_match_star_succeeds_if_tombstone_exists(self):
        self.model.delete_record(self.stored)
        self.validated['header'].pop('If-Match')
        self.validated['header']['If-None-Match'] = '*'
        self.validated['body'] = {'data': {'field': 'new'}}
        self.resource.put()  # not raising.

    def test_post_if_none_match_star_fails_if_record_exists(self):
        self.validated['header'].pop('If-Match')
        self.validated['header']['If-None-Match'] = '*'
        self.resource.request.json = {
            'data': {
                'id': self.stored['id'],
                'field': 'new'}}
        self.validated['body'] = self.resource.request.json
        self.assertRaises(httpexceptions.HTTPPreconditionFailed,
                          self.resource.collection_post)

    def test_post_if_none_match_star_succeeds_if_record_does_not_exist(self):
        self.validated['header'].pop('If-Match')
        self.validated['header']['If-None-Match'] = '*'
        self.validated['body'] = {
            'data': {
                'id': self.resource.model.id_generator(),
                'field': 'new'}}
        self.resource.collection_post()  # not raising.

    def test_get_if_none_match_star_fails_on_collections(self):
        self.validated['header'].pop('If-Match')
        self.validated['header']['If-None-Match'] = '*'
        self.assertRaises(httpexceptions.HTTPPreconditionFailed,
                          self.resource.collection_get)

    def test_delete_if_none_match_star_fails_on_collections(self):
        self.validated['header'].pop('If-Match')
        self.validated['header']['If-None-Match'] = '*'
        self.assertRaises(httpexceptions.HTTPPreconditionFailed,
                          self.resource.collection_delete)

    def test_get_if_match_star_succeeds_if_record_exists(self):
        self.validated['header']['If-Match'] = '*'
        self.resource.record_id = self.stored['id']
        self.resource.get()

    def test_post_if_match_star_succeeds_if_record_exists(self):
        self.validated['header']['If-Match'] = '*'
        self.resource.request.json = {
            'data': {
                'id': self.stored['id'],
                'field': 'new'}}
        self.validated['body'] = self.resource.request.json
        self.resource.collection_post()

    def test_put_if_match_star_succeeds_if_record_exists(self):
        self.validated['header']['If-Match'] = '*'
        self.resource.put()

    def test_patch_if_match_star_succeeds_if_record_exists(self):
        self.validated['header']['If-Match'] = '*'
        self.resource.request.json = {
            'data': {
                'id': self.stored['id'],
                'field': 'new'}}
        self.validated['body'] = self.resource.request.json
        self.resource.patch()

    def test_delete_if_match_star_succeeds_if_record_exists(self):
        self.validated['header']['If-Match'] = '*'
        self.resource.delete()

    def test_put_if_match_star_fails_if_record_does_not_exist(self):
        self.validated['header']['If-Match'] = '*'
        self.resource.request.json = {'data': {'field': 'new'}}
        self.validated['body'] = self.resource.request.json
        self.resource.record_id = self.resource.model.id_generator()
        self.assertRaises(httpexceptions.HTTPPreconditionFailed,
                          self.resource.put)

    def test_put_if_match_current_fails_if_record_does_not_exist(self):
        self.validated['header']['If-Match'] = self.current  # wouldn't raise on existing
        self.resource.request.json = {'data': {'field': 'new'}}
        self.validated['body'] = self.resource.request.json
        self.resource.record_id = self.resource.model.id_generator()
        self.assertRaises(httpexceptions.HTTPPreconditionFailed,
                          self.resource.put)

    def test_put_if_fails_if_if_match_is_superior(self):
        self.validated['header']['If-Match'] += self.current + 10
        self.assertRaises(httpexceptions.HTTPPreconditionFailed,
                          self.resource.put)

    def test_get_if_match_star_suceed_on_collections(self):
        self.validated['header']['If-Match'] = '*'
        self.resource.collection_get()

    def test_delete_if_match_star_suceed_on_collections(self):
        self.validated['header']['If-Match'] = '*'
        self.resource.collection_delete()

    def test_patch_returns_412_if_changed_meanwhile(self):
        self.assertRaises(httpexceptions.HTTPPreconditionFailed,
                          self.resource.patch)

    def test_patch_returns_last_modified_if_changed_meanwhile(self):
        self.resource.timestamp = 0
        try:
            self.resource.patch()
        except httpexceptions.HTTPPreconditionFailed as e:
            error = e
        self.assertNotIn('1970', error.headers['Last-Modified'])

    def test_delete_returns_412_if_changed_meanwhile(self):
        self.assertRaises(httpexceptions.HTTPPreconditionFailed,
                          self.resource.delete)

    def test_delete_all_returns_412_if_changed_meanwhile(self):
        self.assertRaises(httpexceptions.HTTPPreconditionFailed,
                          self.resource.collection_delete)



from kinto.core.testing import unittest

from .support import BaseWebTest


class ContributeViewTest(BaseWebTest, unittest.TestCase):

    def test_returns_info_about_project(self):
        response = self.app.get('/contribute.json')
        keys = sorted(response.json.keys())
        expected = ['description', 'keywords', 'name', 'participate',
                    'repository', 'urls']
        self.assertEqual(keys, expected)



import unittest

from kinto.core.testing import get_user_headers

from .support import (BaseWebTest, MINIMALIST_RECORD,
                      MINIMALIST_GROUP, MINIMALIST_BUCKET,
                      MINIMALIST_COLLECTION)


RECORD_ID = 'd5db6e57-2c10-43e2-96c8-56602ef01435'


class PermissionsViewTest(BaseWebTest, unittest.TestCase):

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['experimental_permissions_endpoint'] = 'True'
        return settings


class EntriesTest(PermissionsViewTest):

    def setUp(self):
        super().setUp()
        self.app.put_json('/buckets/beers', MINIMALIST_BUCKET,
                          headers=self.headers)
        self.app.put_json('/buckets/beers/collections/barley',
                          MINIMALIST_COLLECTION,
                          headers=self.headers)
        self.app.put_json('/buckets/beers/groups/amateurs',
                          MINIMALIST_GROUP,
                          headers=self.headers)
        self.app.put_json('/buckets/beers/collections/barley/records/{}'.format(RECORD_ID),
                          MINIMALIST_RECORD,
                          headers=self.headers)

        # Other user.
        self.app.put_json('/buckets/water', MINIMALIST_BUCKET,
                          headers=get_user_headers('alice'))

    def test_permissions_list_entries_for_current_principal(self):
        resp = self.app.get('/permissions', headers=self.headers)
        permissions = resp.json['data']
        self.assertEqual(len(permissions), 4)

    def test_permissions_can_be_listed_anonymously(self):
        self.app.patch_json('/buckets/beers/collections/barley',
                            {'permissions': {'write': ['system.Everyone']}},
                            headers=self.headers)
        resp = self.app.get('/permissions')
        permissions = resp.json['data']
        self.assertEqual(len(permissions), 1)

    def test_implicit_permissions_are_explicited(self):
        resp = self.app.get('/permissions', headers=get_user_headers('alice'))
        permissions = resp.json['data']
        bucket_permissions = permissions[0]
        self.assertEqual(sorted(bucket_permissions['permissions']),
                         ['collection:create',
                          'group:create',
                          'read',
                          'read:attributes',
                          'write'])

    def test_object_details_are_provided(self):
        resp = self.app.get('/permissions', headers=self.headers)
        entries = resp.json['data']
        for entry in entries:
            if entry['resource_name'] == 'record':
                record_permission = entry
        self.assertEqual(record_permission['record_id'], RECORD_ID)
        self.assertEqual(record_permission['collection_id'], 'barley')
        self.assertEqual(record_permission['bucket_id'], 'beers')

    def test_permissions_list_can_be_filtered(self):
        resp = self.app.get('/permissions?in_resource_name=bucket,collection',
                            headers=self.headers)
        permissions = resp.json['data']
        self.assertEqual(len(permissions), 2)

    def test_filtering_with_unknown_field_is_not_supported(self):
        self.app.get('/permissions?movie=bourne',
                     headers=self.headers,
                     status=400)

    def test_permissions_fields_can_be_selected(self):
        resp = self.app.get('/permissions?_fields=uri',
                            headers=self.headers)
        self.assertNotIn('resource_name', resp.json['data'][0])

    def test_permissions_list_can_be_paginated(self):
        resp = self.app.get('/permissions?_limit=2',
                            headers=self.headers)
        self.assertEqual(resp.headers['Total-Records'], '4')
        self.assertIn('Next-Page', resp.headers)
        self.assertEqual(len(resp.json['data']), 2)

    def test_permissions_list_do_not_crash_with_preconditions(self):
        headers = {'If-None-Match': '"123"', **self.headers}
        self.app.get('/permissions', headers=headers)


class GroupsPermissionTest(PermissionsViewTest):

    def setUp(self):
        super().setUp()

        self.admin_headers = get_user_headers('admin')
        self.admin_principal = self.app.get('/', headers=self.admin_headers).json['user']['id']

        self.app.put_json('/buckets/beers',
                          {'permissions': {'write': ['/buckets/beers/groups/admins']}},
                          headers=self.headers)
        self.app.put_json('/buckets/beers/groups/admins',
                          {'data': {'members': [self.admin_principal]}},
                          headers=self.headers)
        self.app.put_json('/buckets/beers/collections/barley',
                          MINIMALIST_COLLECTION,
                          headers=self.headers)

        self.app.put_json('/buckets/sodas',
                          MINIMALIST_BUCKET,
                          headers=self.headers)
        self.app.put_json('/buckets/beers/groups/admins',
                          {'data': {'members': [self.admin_principal]}},
                          headers=self.headers)
        self.app.put_json('/buckets/sodas/collections/sprite',
                          {'permissions': {'read': ['/buckets/beers/groups/admins']}},
                          headers=self.headers)

    def test_permissions_granted_via_groups_are_listed(self):
        resp = self.app.get('/permissions', headers=self.admin_headers)
        buckets = [e for e in resp.json['data'] if e['resource_name'] == 'bucket']
        self.assertEqual(buckets[0]['id'], 'beers')
        self.assertIn('write', buckets[0]['permissions'])

        collections = [e for e in resp.json['data'] if e['resource_name'] == 'collection']
        self.assertEqual(collections[0]['id'], 'sprite')
        self.assertEqual(collections[0]['bucket_id'], 'sodas')
        self.assertIn('read', collections[0]['permissions'])

    def test_permissions_inherited_are_not_listed(self):
        resp = self.app.get('/permissions', headers=self.admin_headers)
        collections = [e for e in resp.json['data']
                       if e['bucket_id'] == 'beers' and e['resource_name'] == 'collection']
        self.assertEqual(len(collections), 0)


class SettingsPermissionsTest(PermissionsViewTest):

    admin_headers = get_user_headers('admin')
    admin_principal = 'basicauth:bb7fe7b98e759578ef0de85b546dd57d21fe1e399390ad8dafc9886043a00e5c'  # NOQA

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['bucket_write_principals'] = 'system.Authenticated'
        settings['group_create_principals'] = self.admin_principal
        settings['collection_write_principals'] = 'system.Authenticated'
        settings['record_create_principals'] = '/buckets/beers/groups/admins'
        return settings

    def setUp(self):
        super().setUp()
        self.app.put_json('/buckets/beers', MINIMALIST_BUCKET, headers=self.headers)
        self.app.put_json('/buckets/beers/groups/admins',
                          {'data': {'members': [self.admin_principal]}},
                          headers=self.headers)
        self.app.put_json('/buckets/beers/collections/barley',
                          MINIMALIST_COLLECTION,
                          headers=self.headers)

    def test_bucket_write_taken_into_account(self):
        resp = self.app.get('/permissions', headers=get_user_headers("any"))
        buckets = [e for e in resp.json['data'] if e['resource_name'] == 'bucket']
        self.assertEqual(buckets[0]['id'], 'beers')
        self.assertIn('write', buckets[0]['permissions'])

    def test_collection_create_taken_into_account(self):
        resp = self.app.get('/permissions', headers=self.admin_headers)
        buckets = [e for e in resp.json['data'] if e['resource_name'] == 'bucket']
        self.assertEqual(buckets[0]['id'], 'beers')
        self.assertIn('group:create', buckets[0]['permissions'])

    def test_collection_write_taken_into_account(self):
        resp = self.app.get('/permissions', headers=get_user_headers("any"))
        collections = [e for e in resp.json['data'] if e['resource_name'] == 'collection']
        self.assertEqual(collections[0]['id'], 'barley')
        self.assertIn('write', collections[0]['permissions'])

    def test_record_create_taken_into_account(self):
        resp = self.app.get('/permissions', headers=self.admin_headers)
        collections = [e for e in resp.json['data'] if e['resource_name'] == 'collection']
        self.assertEqual(collections[0]['id'], 'barley')
        self.assertIn('record:create', collections[0]['permissions'])

    def test_settings_permissions_are_merged_with_perms_backend(self):
        self.app.patch_json('/buckets/beers',
                            {'permissions': {'collection:create': [self.admin_principal]}},
                            headers=self.headers)
        self.app.patch_json('/buckets/beers/collections/barley',
                            {'permissions': {'read': [self.admin_principal]}},
                            headers=self.headers)

        resp = self.app.get('/permissions', headers=self.admin_headers)

        buckets = [e for e in resp.json['data'] if e['resource_name'] == 'bucket']
        self.assertEqual(buckets[0]['id'], 'beers')
        self.assertIn('group:create', buckets[0]['permissions'])
        self.assertIn('collection:create', buckets[0]['permissions'])

        collections = [e for e in resp.json['data'] if e['resource_name'] == 'collection']
        self.assertEqual(collections[0]['id'], 'barley')
        self.assertIn('record:create', collections[0]['permissions'])
        self.assertIn('read', collections[0]['permissions'])


class DeletedObjectsTest(PermissionsViewTest):

    def setUp(self):
        super().setUp()
        self.app.put_json('/buckets/beers',
                          MINIMALIST_BUCKET,
                          headers=self.headers)
        self.app.put_json('/buckets/beers/groups/admins',
                          MINIMALIST_GROUP,
                          headers=self.headers)
        self.app.put_json('/buckets/beers/collections/barley',
                          MINIMALIST_COLLECTION,
                          headers=self.headers)
        self.app.put_json('/buckets/beers/collections/barley/records/vieuxsinge',
                          MINIMALIST_RECORD,
                          headers=self.headers)
        self.app.delete('/buckets/beers', headers=self.headers)

    def test_deleted_objects_are_not_listed(self):
        resp = self.app.get('/permissions', headers=self.headers)
        self.assertEqual(len(resp.json['data']), 0)



import mock
import unittest
import time

from kinto.core.utils import sqlalchemy
from kinto.core.cache import (CacheBase, memory as memory_backend,
                              postgresql as postgresql_backend)
from kinto.core.cache.testing import CacheTest
from kinto.core.testing import skip_if_no_postgresql


class CacheBaseTest(unittest.TestCase):
    def setUp(self):
        self.cache = CacheBase(cache_prefix='')

    def test_mandatory_overrides(self):
        calls = [
            (self.cache.initialize_schema,),
            (self.cache.flush,),
            (self.cache.ttl, ''),
            (self.cache.expire, '', ''),
            (self.cache.get, ''),
            (self.cache.set, '', ''),
            (self.cache.delete, ''),
        ]
        for call in calls:
            self.assertRaises(NotImplementedError, *call)


class MemoryCacheTest(CacheTest, unittest.TestCase):
    backend = memory_backend
    settings = {
        'cache_prefix': '',
        'cache_max_size_bytes': 7000
    }

    def get_backend_prefix(self, prefix):
        backend_prefix = CacheTest.get_backend_prefix(self, prefix)

        # Share the store between both client for tests.
        backend_prefix._ttl = self.cache._ttl
        backend_prefix._store = self.cache._store

        return backend_prefix

    def test_backend_error_is_raised_anywhere(self):
        pass

    def test_ping_returns_false_if_unavailable(self):
        pass

    def test_ping_logs_error_if_unavailable(self):
        pass

    def test_clean_expired_expires_items(self):
        self.cache.set('foobar', 'toto', 0.01)
        assert 'foobar' in self.cache._store
        assert 'foobar' in self.cache._ttl
        assert 'foobar' in self.cache._created_at
        time.sleep(0.02)
        retrieved = self.cache._clean_expired()
        assert 'foobar' not in self.cache._store
        assert 'foobar' not in self.cache._ttl
        assert 'foobar' not in self.cache._created_at
        self.assertIsNone(retrieved)

    def test_add_over_quota_clean_oversized_items(self):
        for x in range(100):
            # Each entry is 70 bytes
            self.cache.set('foo{0:03d}'.format(x), 'toto')
            time.sleep(0.001)
        assert self.cache.get('foo000') == 'toto'
        # This should delete the 2 first entries
        self.cache.set('foobar', 'tata')
        assert self.cache._quota == 7000 - 70 * 20
        assert self.cache.get('foo000') is None
        assert self.cache.get('foobar') == 'tata'

    def test_size_quota_can_be_set_to_zero(self):
        before = self.cache.max_size_bytes
        self.cache.max_size_bytes = 0
        self.cache.set('foobar', 'tata')
        self.cache.max_size_bytes = before
        assert self.cache.get('foobar') == 'tata'


@skip_if_no_postgresql
class PostgreSQLCacheTest(CacheTest, unittest.TestCase):
    backend = postgresql_backend
    settings = {
        'cache_backend': 'kinto.core.cache.postgresql',
        'cache_pool_size': 10,
        'cache_url': 'postgres://postgres:postgres@localhost:5432/testdb',
        'cache_prefix': ''
    }

    def setUp(self):
        super().setUp()
        self.client_error_patcher = mock.patch.object(
            self.cache.client,
            'session_factory',
            side_effect=sqlalchemy.exc.SQLAlchemyError)



import mock

from pyramid.request import Request

from kinto.core import utils
from kinto.core.authorization import RouteFactory, AuthorizationPolicy, groupfinder
from kinto.core.storage import exceptions as storage_exceptions
from kinto.core.testing import DummyRequest, unittest


class RouteFactoryTest(unittest.TestCase):

    def setUp(self):
        self.record_uri = "/foo/bar"

    def assert_request_resolves_to(self, method, permission, uri=None,
                                   record_not_found=False):
        if uri is None:
            uri = self.record_uri

        with mock.patch('kinto.core.utils.current_service') as current_service:
            # Patch current service.
            resource = mock.MagicMock()
            resource.record_id = 1
            if record_not_found:
                resource.model.get_record.side_effect = \
                    storage_exceptions.RecordNotFoundError
            else:
                resource.model.get_record.return_value = 1
            current_service().resource.return_value = resource

            # Do the actual call.
            request = DummyRequest(method=method)
            request.upath_info = uri
            context = RouteFactory(request)

            self.assertEquals(context.required_permission, permission)

    def test_http_unknown_does_not_raise_a_500(self):
        self.assert_request_resolves_to("unknown", None)

    def test_http_get_resolves_in_a_read_permission(self):
        self.assert_request_resolves_to("get", "read")

    def test_http_post_resolves_in_a_create_permission(self):
        self.assert_request_resolves_to("post", "create")

    def test_http_delete_resolves_in_a_write_permission(self):
        self.assert_request_resolves_to("delete", "write")

    def test_http_put_unexisting_record_resolves_in_a_create_permission(self):
        with mock.patch('kinto.core.utils.current_service') as current_service:
            # Patch current service.
            resource = mock.MagicMock()
            resource.record_id = 1
            resource.model.get_record.side_effect = \
                storage_exceptions.RecordNotFoundError
            current_service().resource.return_value = resource
            current_service().collection_path = '/buckets/{bucket_id}'
            # Do the actual call.
            request = DummyRequest(method='put')
            request.upath_info = '/buckets/abc/collections/1'
            request.matchdict = {'bucket_id': 'abc'}
            context = RouteFactory(request)

            self.assertEquals(context.required_permission, 'create')

    def test_http_put_existing_record_resolves_in_a_write_permission(self):
        self.assert_request_resolves_to("put", "write")

    def test_http_put_sets_current_record_attribute(self):
        with mock.patch('kinto.core.utils.current_service') as current_service:
            # Patch current service.
            resource = mock.MagicMock()
            resource.record_id = 1
            resource.model.get_record.return_value = mock.sentinel.record
            current_service().resource.return_value = resource
            # Do the actual call.
            request = DummyRequest(method='put')
            context = RouteFactory(request)
            self.assertEquals(context.current_record, mock.sentinel.record)

    def test_http_patch_resolves_in_a_write_permission(self):
        self.assert_request_resolves_to("patch", "write")

    def test_attributes_are_none_with_blank_requests(self):
        request = Request.blank(path='/')
        request.registry = mock.Mock(settings={})
        request.authn_type = 'fxa'
        request.prefixed_userid = property(utils.prefixed_userid)
        context = RouteFactory(request)
        self.assertIsNone(context.required_permission)
        self.assertIsNone(context.current_record)
        self.assertIsNone(context.resource_name)

    def test_attributes_are_none_with_non_resource_requests(self):
        basic_service = object()
        request = Request.blank(path='/')
        request.prefixed_userid = property(utils.prefixed_userid)
        request.matched_route = mock.Mock(pattern='foo')
        request.registry = mock.Mock(cornice_services={'foo': basic_service})
        request.registry.settings = {}

        context = RouteFactory(request)
        self.assertIsNone(context.current_record)
        self.assertIsNone(context.required_permission)
        self.assertIsNone(context.resource_name)

    def test_fetch_shared_records_uses_pattern_if_on_collection(self):
        request = DummyRequest()
        request.route_path.return_value = '/v1/buckets/%2A'
        service = mock.MagicMock()
        service.type = 'collection'
        with mock.patch('kinto.core.authorization.utils.current_service') as m:
            m.return_value = service
            context = RouteFactory(request)
        self.assertTrue(context.on_collection)

        context.fetch_shared_records('read', ['userid'], None)

        request.registry.permission.get_accessible_objects.assert_called_with(
            ['userid'],
            [('/buckets/*', 'read')],
            with_children=False)

    def test_fetch_shared_records_uses_get_bound_permission_callback(self):
        request = DummyRequest()
        service = mock.MagicMock()
        request.route_path.return_value = '/v1/buckets/%2A'
        service.type = 'collection'
        with mock.patch('kinto.core.authorization.utils.current_service') as m:
            m.return_value = service
            context = RouteFactory(request)
        self.assertTrue(context.on_collection)

        # Define a callback where write means read:
        def get_bound_perms(obj_id, perm):
            return [(obj_id, 'write'), (obj_id, 'read')]

        context.fetch_shared_records('read', ['userid'], get_bound_perms)

        request.registry.permission.get_accessible_objects.assert_called_with(
            ['userid'],
            [('/buckets/*', 'write'), ('/buckets/*', 'read')],
            with_children=False)

    def test_fetch_shared_records_sets_shared_ids_from_results(self):
        request = DummyRequest()
        context = RouteFactory(request)
        request.registry.permission.get_accessible_objects.return_value = {
            '/obj/1': ['read', 'write'],
            '/obj/3': ['obj:create']
        }
        context.fetch_shared_records('read', ['userid'], None)
        self.assertEquals(sorted(context.shared_ids), ['1', '3'])

    def test_fetch_shared_records_sets_shared_ids_if_empty(self):
        request = DummyRequest()
        context = RouteFactory(request)
        request.registry.permission.get_accessible_objects.return_value = {}

        context.fetch_shared_records('read', ['userid'], None)

        self.assertEqual(context.shared_ids, [])


class AuthorizationPolicyTest(unittest.TestCase):
    def setUp(self):
        self.authz = AuthorizationPolicy()
        self.context = mock.MagicMock()
        self.context.permission_object_id = '/articles/43/comments/2'
        self.context.required_permission = 'read'
        self.principals = []
        self.context.get_prefixed_principals.return_value = self.principals
        self.permission = 'dynamic'

    def test_permits_does_not_refer_to_context_if_permission_is_private(self):
        self.assertFalse(self.authz.permits(None, [], 'private'))

    def test_permits_return_if_authenticated_when_permission_is_private(self):
        self.assertTrue(self.authz.permits(None,
                                           ['system.Authenticated'],
                                           'private'))

    def test_permits_refers_to_context_to_check_permissions(self):
        self.context.check_permission.return_value = True
        allowed = self.authz.permits(self.context, self.principals, 'dynamic')
        self.assertTrue(allowed)

    def test_permits_refers_to_context_to_check_permission_principals(self):
        self.context.check_permission.return_value = False
        allowed = self.authz.permits(
            self.context, ['fxa:user', 'system.Authenticated'], 'dynamic')
        self.assertTrue(allowed)

    def test_permits_reads_the_context_when_permission_is_dynamic(self):
        self.authz.permits(self.context, self.principals, 'dynamic')
        self.context.check_permission.assert_called_with(
            self.principals,
            [(self.context.permission_object_id, 'read')])

    def test_permits_uses_get_bound_permissions_if_defined(self):
        self.authz.get_bound_permissions = lambda o, p: mock.sentinel.callback
        self.authz.permits(self.context, self.principals, 'dynamic')
        self.context.check_permission.assert_called_with(
            self.principals,
            mock.sentinel.callback)

    def test_permits_calls_get_bound_permissions_with_context_info(self):
        self.authz.get_bound_permissions = mock.Mock(return_value=[])
        self.authz.permits(self.context, self.principals, 'dynamic')
        self.authz.get_bound_permissions.assert_called_with(
            self.context.permission_object_id,
            'read')

    def test_permits_consider_permission_when_not_dynamic(self):
        self.authz.permits(self.context, self.principals, 'foobar')
        self.context.check_permission.assert_called_with(
            self.principals,
            [(self.context.permission_object_id, 'foobar')])

    def test_permits_prepend_obj_type_to_permission_on_create(self):
        self.context.required_permission = 'create'
        self.context.resource_name = 'record'
        self.authz.permits(self.context, self.principals, 'dynamic')
        self.context.check_permission.assert_called_with(
            self.principals,
            [(self.context.permission_object_id, 'record:create')])

    def test_permits_takes_route_factory_allowed_principals_into_account(self):
        self.context.resource_name = 'record'
        self.context.required_permission = 'create'
        self.context._settings = {'record_create_principals': 'fxa:user'}
        allowed = self.authz.permits(self.context, ['fxa:user'], 'dynamic')
        self.context._check_permission.assert_not_called()
        self.assertTrue(allowed)


class GuestAuthorizationPolicyTest(unittest.TestCase):
    def setUp(self):
        self.authz = AuthorizationPolicy()
        self.authz.get_bound_permissions = lambda o, p: []
        self.request = DummyRequest(method='GET')
        self.context = RouteFactory(self.request)
        self.context.on_collection = True
        self.context.check_permission = mock.Mock(return_value=False)

    def test_permits_returns_true_if_collection_and_shared_records(self):
        self.context.fetch_shared_records = mock.MagicMock(return_value=[
            'record1', 'record2'])
        allowed = self.authz.permits(self.context, ['userid'], 'dynamic')
        # Note: we use the list of principals from request.prefixed_principals
        self.context.fetch_shared_records.assert_called_with(
            'read',
            ['system.Everyone', 'system.Authenticated', 'basicauth:bob'],
            self.authz.get_bound_permissions)
        self.assertTrue(allowed)

    def test_permits_does_not_return_true_if_not_collection(self):
        self.context.on_collection = False
        allowed = self.authz.permits(self.context, ['userid'], 'dynamic')
        self.assertFalse(allowed)

    def test_permits_does_not_return_true_if_not_list_operation(self):
        self.context.required_permission = 'create'
        allowed = self.authz.permits(self.context, ['userid'], 'dynamic')
        self.assertFalse(allowed)
        allowed = self.authz.permits(self.context, ['userid'], 'create')
        self.assertFalse(allowed)

    def test_permits_returns_false_if_collection_is_unknown(self):
        self.context.fetch_shared_records = mock.MagicMock(return_value=None)
        allowed = self.authz.permits(self.context, ['userid'], 'dynamic')
        # Note: we use the list of principals from request.prefixed_principals
        self.context.fetch_shared_records.assert_called_with(
            'read',
            ['system.Everyone', 'system.Authenticated', 'basicauth:bob'],
            self.authz.get_bound_permissions)
        self.assertFalse(allowed)

    def test_perm_object_id_is_naive_if_no_record_path_exists(self):
        def route_path(service_name, **kwargs):
            # Simulate a resource that has no record_path (only list).
            if service_name == 'article-record':
                raise KeyError
            return '/comments/sub/{id}'.format_map(kwargs)

        self.request.route_path.side_effect = route_path

        self.request.path = '/comments'
        self.context.resource_name = 'comment'
        obj_id = self.context.get_permission_object_id(self.request, '*')
        self.assertEquals(obj_id, '/comments/sub/*')

        self.request.path = '/articles'
        self.context.resource_name = 'article'
        obj_id = self.context.get_permission_object_id(self.request, '*')
        self.assertEquals(obj_id, '/articles/*')


class GroupFinderTest(unittest.TestCase):
    def setUp(self):
        self.request = DummyRequest(method='GET')

    def test_uses_prefixed_as_userid(self):
        self.request.prefixed_userid = 'basic:bob'
        groupfinder('bob', self.request)
        self.request.registry.permission.get_user_principals.assert_called_with('basic:bob')

    def test_uses_provided_id_if_no_prefixed_userid(self):
        self.request.prefixed_userid = None
        groupfinder('bob', self.request)
        self.request.registry.permission.get_user_principals.assert_called_with('bob')



from kinto import logger
from kinto.core.cache import CacheBase
from kinto.core.utils import msec_time
from kinto.core.decorators import synchronized


class Cache(CacheBase):
    """Cache backend implementation in local process memory.

    Enable in configuration::

        kinto.cache_backend = kinto.core.cache.memory

    :noindex:
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.flush()

    def initialize_schema(self, dry_run=False):
        # Nothing to do.
        pass

    def flush(self):
        self._created_at = {}
        self._ttl = {}
        self._store = {}
        self._quota = 0

    def _clean_expired(self):
        current = msec_time()
        expired = [k for k, v in self._ttl.items() if current >= v]
        for expired_item_key in expired:
            self.delete(expired_item_key[len(self.prefix):])

    def _clean_oversized(self):
        if self._quota < self.max_size_bytes:
            return

        for key, value in sorted(self._created_at.items(), key=lambda k: k[1]):
            if self._quota < (self.max_size_bytes * 0.8):
                break
            self.delete(key[len(self.prefix):])

    @synchronized
    def ttl(self, key):
        ttl = self._ttl.get(self.prefix + key)
        if ttl is not None:
            return (ttl - msec_time()) / 1000.0
        return -1

    @synchronized
    def expire(self, key, ttl):
        self._ttl[self.prefix + key] = msec_time() + int(ttl * 1000.0)

    @synchronized
    def set(self, key, value, ttl=None):
        self._clean_expired()
        self._clean_oversized()
        if ttl is not None:
            self.expire(key, ttl)
        else:
            logger.warning("No TTL for cache key '{}'".format(key))
        item_key = self.prefix + key
        self._store[item_key] = value
        self._created_at[item_key] = msec_time()
        self._quota += size_of(item_key, value)

    @synchronized
    def get(self, key):
        self._clean_expired()
        return self._store.get(self.prefix + key)

    @synchronized
    def delete(self, key):
        key = self.prefix + key
        self._ttl.pop(key, None)
        self._created_at.pop(key, None)
        value = self._store.pop(key, None)
        self._quota -= size_of(key, value)


def load_from_config(config):
    settings = config.get_settings()
    return Cache(cache_prefix=settings['cache_prefix'],
                 cache_max_size_bytes=settings['cache_max_size_bytes'])


def size_of(key, value):
    # Key used for ttl, created_at and store.
    # Int size is 24 bytes one for ttl and one for created_at values
    return len(key) * 3 + len(str(value)) + 24 * 2



import colander
from pyramid.security import NO_PERMISSION_REQUIRED, Authenticated

from kinto.core import Service

hello = Service(name="hello", path='/', description="Welcome")


class HelloResponseSchema(colander.MappingSchema):
    body = colander.SchemaNode(colander.Mapping(unknown='preserve'))


hello_response_schemas = {
    '200': HelloResponseSchema(
        description='Return information about the running Instance.')
}


@hello.get(permission=NO_PERMISSION_REQUIRED, tags=['Utilities'],
           operation_id='server_info', response_schemas=hello_response_schemas)
def get_hello(request):
    """Return information regarding the current instance."""
    settings = request.registry.settings
    project_name = settings['project_name']
    project_version = settings['project_version']
    data = dict(
        project_name=project_name,
        project_version=project_version,
        http_api_version=settings['http_api_version'],
        project_docs=settings['project_docs'],
        url=request.route_url(hello.name)
    )

    eos = get_eos(request)
    if eos:
        data['eos'] = eos

    data['settings'] = {}
    public_settings = request.registry.public_settings
    for setting in list(public_settings):
        data['settings'][setting] = settings[setting]

    # If current user is authenticated, add user info:
    # (Note: this will call authenticated_userid() with multiauth+groupfinder)
    if Authenticated in request.effective_principals:
        data['user'] = request.get_user_info()

    # Application can register and expose arbitrary capabilities.
    data['capabilities'] = request.registry.api_capabilities

    return data


def get_eos(request):
    return request.registry.settings['eos']



from kinto.core import Service
from kinto.core.authorization import RouteFactory


attachment = Service(name='attachment',
                     description='Attach file to record',
                     path='/attachment',
                     factory=RouteFactory)


@attachment.post(permission='attach')
def attachment_post(request):
    return {"ok": True}


def includeme(config):
    config.add_cornice_service(attachment)
